import datetime
import errno
import functools
import logging
import os.path
import re
import socket
import sys
import threading
import time
from hashlib import sha256
from urllib.parse import unquote_to_bytes, quote

import crcmod as crcmod
import pytz
import six
from pytz import unicode
from urllib3.util.connection import allowed_gai_family

from .consts import (DEFAULT_MIMETYPE, GMT_DATE_FORMAT,
                     MAX_PART_NUMBER, MAX_PART_SIZE, MIN_PART_SIZE, CHUNK_SIZE,
                     CLIENT_ENCRYPTION_ALGORITHM, SERVER_ENCRYPTION_ALGORITHM, LAST_MODIFY_TIME_DATE_FORMAT,
                     EMPTY_SHA256_HASH, PAYLOAD_BUFFER)
from .enum import DataTransferType, ACLType, StorageClassType, MetadataDirectiveType, AzRedundancyType, PermissionType, \
    GranteeType, CannedType
from .exceptions import TosClientError
from .mine_type import TYPES_MAP

logger = logging.getLogger(__name__)
REGION_MAP = {
    'cn-beijing': 'tos-cn-beijing.volces.com',
    'cn-guangzhou': 'tos-cn-guangzhou.volces.com',
    'cn-shanghai': 'tos-cn-shanghai.volces.com',
    'ap-southeast-1': 'tos-ap-southeast-1.volces.com',
}

S3_REGION_LIST = ['tos-s3-cn-beijing.volces.com',
                  'tos-s3-cn-guangzhou.volces.com',
                  'tos-s3-cn-shanghai.volces.com',
                  'tos-s3-ap-southeast-1.volces.com']


def get_value(kv, key, handler=lambda x: x):
    if key in kv:
        return handler(kv[key])
    else:
        return None


def get_etag(kv):
    return get_value(kv, "ETag", lambda x: x.strip('"'))


def to_bytes(data):
    if six.PY2:
        if isinstance(data, unicode):
            return data.encode('utf-8')
    if six.PY3:
        if isinstance(data, str):
            return data.encode(encoding='utf-8')
    return data


def to_str(data):
    if six.PY2:
        if isinstance(data, unicode):
            return data.encode('utf-8')
    if six.PY3:
        if isinstance(data, bytes):
            return data.decode('utf-8')
    return data


def parse_modify_time_to_utc_datetime(value):
    return datetime.datetime.strptime(value, LAST_MODIFY_TIME_DATE_FORMAT).replace(tzinfo=pytz.utc)


def parse_gmt_time_to_utc_datetime(value):
    return datetime.datetime.strptime(value, GMT_DATE_FORMAT).replace(tzinfo=pytz.utc)


def get_content_type(key):
    """根据文件名后缀，获取文件类型"""
    ext = os.path.splitext(key)[1].lower()
    return TYPES_MAP[ext] if ext in TYPES_MAP else DEFAULT_MIMETYPE


def init_path(path: str, key: str):
    try_make_file_dir(path)
    if os.path.isdir(path) and key.endswith('/'):
        return True
    return False


def try_make_file_dir(path: str):
    if os.path.isdir(path) or os.path.isfile(path):
        return
    file_dir, file_name = os.path.split(path)
    if file_dir:
        os.makedirs(file_dir, exist_ok=True)


def init_checkpoint_dir(checkpoint_file: str):
    if os.path.isdir(checkpoint_file):
        return checkpoint_file
    file_dir, file_name = os.path.split(checkpoint_file)
    if file_dir:
        os.makedirs(file_dir, exist_ok=True)
        return file_dir
    raise TosClientError('checkpoint_file is invalid')


def _make_range_string(start, last):
    if start is None and last is None:
        return ''

    if start is not None and last is not None and start > last:
        raise TosClientError('invalid range format')

    return 'bytes=' + _range(start, last)


def _range(start, last):
    def to_str(pos):
        if pos is None:
            return ''
        else:
            return str(pos)

    return to_str(start) + '-' + to_str(last)


def to_unicode(data):
    """把输入转换为unicode，要求输入是unicode或者utf-8编码的bytes。"""
    return to_str(data)


def is_utf8_with_trigger(s: bytes):
    i = 0
    while i < len(s):
        now = s[i]
        first = first_zero(s[i])
        if first == 0:
            if s[i] < 32 or s[i] == 127:
                return False
            i = i + 1
        elif 2 <= first <= 6:
            if i + first <= len(s):
                for j in range(i + 1, i + first):
                    if s[j] & 0xC0 != 0x80:
                        return False
                i += first
            else:
                return False
        else:
            return False
    return True


def first_zero(a):
    mask = 0x80
    i = 0
    while mask & a != 0:
        mask = mask >> 1
        i = i + 1
    return i


def generate_http_proxies(ip: str, port: int, user_name: str = None, password: str = None):
    # 未配置ip地址 返回空{}
    if ip is None or len(ip) == 0:
        return {}

    proxy = ""
    if user_name is None or len(user_name) == 0:
        # 返回 "http": "http://{ip}:port",
        proxy = 'http://{0}:{1}'.format(ip, port)

    else:
        # 返回  "http": "http://{user_name}:{password}@{ip}:{port}/"
        proxy = "http://{0}:{1}@{2}:{3}/".format(user_name, password, ip, port)

    return {'http': proxy}


def _make_copy_source(src_bucket, src_key, src_version_id):
    copy_source = {}
    if src_bucket:
        copy_source['Bucket'] = src_bucket

    if src_key:
        copy_source['Key'] = src_key

    if src_version_id:
        copy_source['VersionId'] = src_version_id

    return copy_source


def _make_upload_part_file_content(file, offset, part_size, size):
    """创建 具备offset 和读取长度 file-object
    """
    if offset < 0 or part_size < -1:
        return None
    if offset > size:
        return None

    file.seek(offset, os.SEEK_SET)

    if part_size == -1 or (part_size + offset) >= size:
        return SizeAdapter(file, size - offset, init_offset=offset, can_reset=True)
    else:
        return SizeAdapter(file, part_size, init_offset=offset, can_reset=True)


class SizeAdapter(object):
    def __init__(self, file_object, size, init_offset=None, can_reset=False):
        self.file_object = file_object
        self.size = size
        self.offset = 0
        self.init_offset = init_offset
        self.can_reset = can_reset
        if init_offset:
            self.file_object.seek(init_offset, os.SEEK_SET)

    def read(self, amt=None):
        if self.offset >= self.size:
            return ''

        if (amt is None or amt < 0) or (amt + self.offset >= self.size):
            data = self.file_object.read(self.size - self.offset)
            self.offset = self.size
            return data

        self.offset += amt
        return self.file_object.read(amt)

    @property
    def len(self):
        return self.size

    def reset(self):
        if self.can_reset:
            self.offset = 0
            if self.init_offset is not None:
                self.file_object.seek(self.init_offset, os.SEEK_SET)


def meta_header_encode(query, doseq=False, safe='', encoding=None, errors=None,
                       quote_via=quote):
    headers = {}
    if hasattr(query, "items"):
        query = query.items()
    else:
        # It's a bother at times that strings and string-like objects are
        # sequences.
        try:
            # non-sequence items should not work with len()
            # non-empty strings will fail this
            if len(query) and not isinstance(query[0], tuple):
                raise TypeError
            # Zero-length sequences of all types will get here and succeed,
            # but that's a minor nit.  Since the original implementation
            # allowed empty dicts that type of behavior probably should be
            # preserved for consistency
        except TypeError:
            ty, va, tb = sys.exc_info()
            raise TypeError("not a valid non-string sequence "
                            "or mapping object").with_traceback(tb)

    if not doseq:
        for k, v in query:
            if isinstance(k, bytes):
                k = quote_via(k, safe)
            else:
                k = quote_via(str(k), safe, encoding, errors)

            if isinstance(v, bytes):
                v = quote_via(v, safe)
            else:
                v = quote_via(str(v), safe, encoding, errors)
            headers[k] = v
    else:
        for k, v in query:
            if isinstance(k, bytes):
                k = quote_via(k, safe)
            else:
                k = quote_via(str(k), safe, encoding, errors)

            if isinstance(v, bytes):
                v = quote_via(v, safe)
                headers[k] = v
            elif isinstance(v, str):
                v = quote_via(v, safe, encoding, errors)
                headers[k] = v
            else:
                try:
                    # Is this a sufficient test for sequence-ness?
                    len(v)
                except TypeError:
                    # not a sequence
                    v = quote_via(str(v), safe, encoding, errors)
                    headers[k] = v
                else:
                    # loop over the sequence
                    for elt in v:
                        if isinstance(elt, bytes):
                            elt = quote_via(elt, safe)
                        else:
                            elt = quote_via(str(elt), safe, encoding, errors)
                        headers[k] = v
    return headers


def meta_header_decode(headers):
    decode_headers = {}
    encoding = 'utf-8'
    errors = 'replace'

    for en in headers:
        k = en
        v = headers[en]
        if '%' in k:
            k = unquote_to_bytes(en).decode(encoding, errors)
        if '%' in v:
            v = unquote_to_bytes(headers[en]).decode(encoding, errors)
        decode_headers[k] = v
    return decode_headers


def makedir_p(dirpath):
    try:
        os.makedirs(dirpath)
    except os.error as e:
        if e.errno != errno.EEXIST:
            raise


def get_number(m, n):
    return (m + n - 1) // n


def file_bytes_to_read(fileobj):
    current = fileobj.tell()

    fileobj.seek(0, os.SEEK_END)
    end = fileobj.tell()
    fileobj.seek(current, os.SEEK_SET)

    return end - current


def get_parent_directory_from_File(absolute_path_to_file):
    arr = absolute_path_to_file.split(os.sep)[:-1]
    return os.sep.join(arr)


def _cal_progress_callback(progress_callback, consumed_bytes, total_bytes, rw_once_bytes,
                           type: DataTransferType):
    if progress_callback:
        progress_callback(consumed_bytes, total_bytes, rw_once_bytes, type)


def _cal_rate_limiter_callback(rate_limiter, want):
    if rate_limiter:
        while True:
            result = rate_limiter.acquire(want)
            if result.ok:
                return
            time.sleep(result.time_to_wait)


def _cal_crc_callback(crc_callback, content, discard=0):
    if crc_callback:
        crc_callback(content[discard:])


def BaseCallback(object):
    def success():
        pass

    def fail():
        pass

    def abort():
        pass


class _ReaderAdapter(object):
    """
    通过adapter模式，实现进度条监控、客户端限速、上传crc计算
    """

    def __init__(self, data, progress_callback=None, size=None, crc_callback=None,
                 limiter_callback=None, download_operator=False, can_reset=False,
                 init_offset=None):
        self.data = to_bytes(data)
        self.progress_callback = progress_callback
        self.size = size
        self.offset = 0
        self.download_operator = download_operator

        self.crc_callback = crc_callback
        self.limiter_callback = limiter_callback
        if hasattr(data, 'can_reset'):
            self.can_reset = data.can_reset
        else:
            self.can_reset = can_reset
        self.init_offset = init_offset
        if self.init_offset is not None:
            self.data.seek(init_offset, os.SEEK_SET)

    @property
    def len(self):
        return self.size

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        content = self.read(CHUNK_SIZE)
        if content:
            return content
        else:
            raise StopIteration

    def read(self, amt=None):
        if self.offset >= self.size:
            _cal_progress_callback(self.get_progress(), self.size, self.size, 0,
                                   DataTransferType.Data_Transfer_Succeed)
            return to_bytes('')
        if self.offset == 0:
            _cal_progress_callback(self.progress_callback, min(self.offset, self.size), self.size, 0,
                                   DataTransferType.Data_Transfer_Started)
        if amt is None or amt < 0:
            bytes_to_read = self.size - self.offset
        else:
            bytes_to_read = min(amt, self.size - self.offset)

        if isinstance(self.data, bytes):
            content = self.data[self.offset:self.offset + bytes_to_read]
        else:
            content = to_bytes(self.data.read(bytes_to_read))

        self.offset += bytes_to_read

        _cal_progress_callback(self.progress_callback, min(self.offset, self.size), self.size, bytes_to_read,
                               DataTransferType.Data_Transfer_RW)

        _cal_crc_callback(self.crc_callback, content)

        _cal_rate_limiter_callback(self.limiter_callback, bytes_to_read)
        return content

    @property
    def crc(self):
        if self.crc_callback:
            return self.crc_callback.crc
        else:
            return 0

    def reset(self):
        if self.can_reset and self.size != -1:
            self.offset = 0
            if self.crc_callback:
                self.crc_callback = self.crc_callback.reset()
            if self.init_offset is not None:
                self.data.seek(self.init_offset, os.SEEK_SET)
            if isinstance(self.data, _ReaderAdapter) or isinstance(self.data, SizeAdapter):
                self.data.reset()

    def get_progress(self):
        if self.progress_callback:
            return self.progress_callback
        elif isinstance(self.data, _ReaderAdapter):
            return self.data.get_progress()
        return None


# 只对上传时会将网络流包或具备__iter__装成 _IterableAdapter 适配器
# 不支持超时重试、进度条功能
class _IterableAdapter(object):
    def __init__(self, data, progress_callback=None, crc_callback=None, limiter_callback=None, download_operator=False,
                 can_reset=False):
        self.iter = iter(data)
        self.progress_callback = progress_callback
        self.crc_callback = crc_callback
        self.limiter_callback = limiter_callback
        self.download_operator = download_operator
        self.can_reset = False
        self.offset = 0

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        if self.offset == 0:
            _cal_progress_callback(self.progress_callback, self.offset, -1, 0,
                                   DataTransferType.Data_Transfer_Started)
        try:
            content = next(self.iter)
        except Exception as e:
            _cal_progress_callback(self.progress_callback, self.offset, self.offset, 0,
                                   DataTransferType.Data_Transfer_Succeed)
            raise e
        size = len(content)
        _cal_progress_callback(self.progress_callback, self.offset, -1, size,
                               DataTransferType.Data_Transfer_RW)

        _cal_crc_callback(self.crc_callback, content)

        _cal_rate_limiter_callback(self.limiter_callback, size)
        self.offset = self.offset + size
        return content

    def get_progress(self):
        if self.progress_callback:
            return self.progress_callback
        elif isinstance(self.iter, _ReaderAdapter):
            return self.iter.get_progress()
        return None

    @property
    def crc(self):
        if self.crc_callback:
            return self.crc_callback.crc
        else:
            return 0


from .thread_ctx import produce_body


def patch_content(content):
    if isinstance(content, _ReaderAdapter) or (hasattr(content, 'read') and callable(content.read)):
        _read = content.read

        def read(amt=None):
            data = _read(amt)
            produce_body(len(data))
            return data

        content.read = read
    elif isinstance(content, _IterableAdapter):
        _next = content.next

        def next():
            data = _next()
            produce_body(len(data))
            return data

        content.next = next


def init_content(data, can_reset=None, init_offset=None):
    """ 此方法用于将上传数据统一包装为 _ReaderAdapter 或 _IterableAdapter，方便后续统一方式处理超时重试

    @param data: 待包装数据
    @param can_reset: 是否可reset
    @param init_offset: 初始数据偏移
    @return:
    """

    # 说明为 put_object_from_file等方式，以及对流进行包装，调用put_object方式时，无需包装
    if isinstance(data, _ReaderAdapter) or isinstance(data, _IterableAdapter):
        return data

    # 明确可以reset
    if can_reset is True:
        return add_Background_func(data, can_reset=True, init_offset=init_offset)

    # 具备seek、tell方法, 初步判定可以 reset
    if hasattr(data, 'seek') and hasattr(data, 'tell') and init_offset is not None:
        return add_Background_func(data, can_reset=True, init_offset=init_offset)

    # 为SizeAdapter及使用限制结束位的文件上传方式，可reset
    if isinstance(data, SizeAdapter):
        return add_Background_func(data, can_reset=True)

    # 及为非IO对象，(str bytes)等初步判定可reset
    if not (hasattr(data, 'seek') and hasattr(data, 'tell')):
        return add_Background_func(data, can_reset=True)

    # 兜底不可reset
    return add_Background_func(data, can_reset=False)


def add_Background_func(data, can_reset=False, init_offset=None, size=None):
    """
    此方法根据数据的data类型，将date包装为 _IterableAdapter 或 _ReaderAdapter
    1. 尝试将数据转化为bytes
    2. 尝试获取数据的size(len、seek、tell)
        1. 网络流具备seek、tell但未实现因此返回为None
    3. size 为空 但具备 __iter__ 直接封装为_IterableAdapter、通过http chuck方式发送
    4. size 不为空，直接封装为 _ReaderAdapter
    """
    data = to_bytes(data)
    if size is None:
        size = _get_size(data)
    if size is None and hasattr(data, '__iter__'):
        return _IterableAdapter(data, can_reset=False)
    elif size is not None:
        return _ReaderAdapter(data, size=size, can_reset=can_reset, init_offset=init_offset)
    raise TosClientError('{0} is not a file object, nor an iterator'.format(data.__class__.__name__))


def add_progress_listener_func(data, progress_callback, download_operator=False, can_reset=False,
                               init_offset=None, size=None, is_response=False):
    """
    向data添加进度条监控功能，通过adapter模式实现
    """
    data = to_bytes(data)
    if size is None:
        size = _get_size(data)
    if size is None and hasattr(data, '__iter__'):
        return _IterableAdapter(data, progress_callback=progress_callback, can_reset=False)
    elif size is not None or is_response:
        return _ReaderAdapter(data=data, progress_callback=progress_callback, size=size,
                              download_operator=download_operator, can_reset=can_reset, init_offset=init_offset)
    raise TosClientError('{0} is not a file object, nor an iterator'.format(data.__class__.__name__))


def add_rate_limiter_func(data, rate_limiter, size=None, can_reset=False, init_offset=None, is_response=False):
    """
    返回一个适配器，从而在读取、上传 'data'时，能够通过令牌桶算法进行限速
    """
    data = to_bytes(data)
    if size is None:
        size = _get_size(data)
    if size is None and hasattr(data, '__iter__'):
        return _IterableAdapter(data, limiter_callback=rate_limiter, can_reset=False)
    elif size is not None or is_response:
        return _ReaderAdapter(data=data, limiter_callback=rate_limiter, size=size, can_reset=can_reset,
                              init_offset=init_offset)
    raise TosClientError('{0} is not a file object, nor an iterator'.format(data.__class__.__name__))


def add_crc_func(data, init_crc=0, discard=0, size=None, can_reset=False, is_response=False):
    """
    向data中添加crc计算功能，实现上传和下载对象后得到本地对象crc计算值
    """
    data = to_bytes(data)
    if size is None:
        size = _get_size(data)
    if size is None and hasattr(data, '__iter__'):
        return _IterableAdapter(data, crc_callback=Crc64(init_crc), can_reset=False)
    elif size is not None or is_response:
        return _ReaderAdapter(data, size=size, crc_callback=Crc64(init_crc), can_reset=can_reset)
    raise TosClientError('{0} is not a file object, nor an iterator'.format(data.__class__.__name__))


def _get_size(data):
    if hasattr(data, '__len__') or hasattr(data, 'len') or (hasattr(data, 'seek') and hasattr(data, 'tell')) or hasattr(
            data, 'read'):
        if hasattr(data, '__len__'):
            return len(data)

        if hasattr(data, 'len'):
            return data.len

        # 具备 seek 和 tell 方法有可能为文件，也可能为网络流， 因此需调用 file_byte_to_read 方法尝试读取长度
        # 抛出异常说明为网络流
        try:
            if hasattr(data, 'seek') and hasattr(data, 'tell'):
                return file_bytes_to_read(data)
        except:
            return None
    return None


class TokenBucketResult(object):
    def __init__(self, ok: bool, time_to_wait: int):
        self.ok = ok
        self.time_to_wait = time_to_wait


class RateLimiter(object):
    """
    令牌桶算法实现
    """

    # rate是令牌发放速度，capacity是桶的⼤⼩
    def __init__(self, rate, capacity):
        self._rate = rate
        self._capacity = capacity
        self._current_amount = 0
        self._last_consume_time = int(time.time())
        self._lock = threading.Lock()

    # want是发送数据需要的令牌数
    def acquire(self, want):
        with self._lock:
            increment = (int(time.time()) - self._last_consume_time) * self._rate  # 计算从上次发送到这次发送，新发放的令牌数量
            self._current_amount = min(
                increment + self._current_amount, self._capacity)  # 令牌数量不能超过桶的容量
            if want > self._current_amount:  # 如果没有⾜够的令牌，则不能发送数据
                time_to_wait = (want - self._current_amount) / self._rate
                return TokenBucketResult(False, int(time_to_wait))
            self._last_consume_time = int(time.time())
            self._current_amount -= want
            return TokenBucketResult(True, 0)

    def reset(self):
        r = RateLimiter(self._rate, self._capacity)
        return r


class Crc64(object):
    _POLY = 0x142F0E1EBA9EA3693
    _XOROUT = 0XFFFFFFFFFFFFFFFF

    def __init__(self, init_crc=0):
        init_crc = int(init_crc)
        self.init_crc = init_crc
        self.crc64 = crcmod.Crc(self._POLY, initCrc=init_crc, rev=True, xorOut=self._XOROUT)

        self.crc64_combineFun = mkCombineFun(self._POLY, initCrc=init_crc, rev=True, xorOut=self._XOROUT)

    def __call__(self, data):
        self.update(data)

    def update(self, data):
        self.crc64.update(data)

    def combine(self, crc1, crc2, len2):
        crc2 = int(crc2)
        return self.crc64_combineFun(crc1, crc2, len2)

    @property
    def crc(self):
        return self.crc64.crcValue

    def reset(self):
        c = Crc64(init_crc=self.init_crc)
        return c


def check_crc(operation, client_crc, tos_crc, request_id):
    if tos_crc is not None:
        tos_crc = int(tos_crc)
    if client_crc != tos_crc:
        raise TosClientError(
            "Check CRC failed: req_id: {0}, operation: {1}, CRC checksum of client: {2} is mismatch "
            "with tos: {3}".format(request_id, operation, client_crc, tos_crc))


def rename_file(src, dst):
    os.rename(src, dst)


def cal_crc_from_download_parts(parts, init_crc=0):
    client_crc = 0
    crc_obj = Crc64(init_crc)
    for part in parts:
        if not part.part_crc or not part.size:
            return None
        client_crc = crc_obj.combine(client_crc, part.part_crc, part.size)
    return client_crc


def cal_crc_from_upload_parts(parts, init_crc=0):
    client_crc = 0
    crc_obj = Crc64(init_crc)
    for part in parts:
        client_crc = crc_obj.combine(client_crc, part.hash_crc64_ecma, part.part_size)
    return client_crc


def copy_and_verify_length(src, dst, expected_len,
                           chunk_size=16 * 1024,
                           request_id=''):
    num = 0

    while 1:
        buf = src.read(chunk_size)
        if not buf:
            break

        num += len(buf)
        dst.write(buf)

    if num != expected_len:
        raise TosClientError("Some error from read source, request_id:{0}".format(request_id))


is_py3 = (sys.version_info[0] == 3)
if is_py3:
    xrange = range
    long = int
    sys.maxint = sys.maxsize


def mkCombineFun(poly, initCrc=~long(0), rev=True, xorOut=0):
    (sizeBits, initCrc, xorOut) = _verifyParams(poly, initCrc, xorOut)

    mask = (long(1) << sizeBits) - 1
    if rev:
        poly = _bitrev(long(poly) & mask, sizeBits)
    else:
        poly = long(poly) & mask

    if sizeBits == 64:
        fun = _combine64
    else:
        raise NotImplementedError

    def combine_fun(crc1, crc2, len2):
        return fun(poly, initCrc ^ xorOut, rev, xorOut, crc1, crc2, len2)

    return combine_fun


GF2_DIM = 64


def gf2_matrix_square(square, mat):
    for n in xrange(GF2_DIM):
        square[n] = gf2_matrix_times(mat, mat[n])


def gf2_matrix_times(mat, vec):
    summary = 0
    mat_index = 0

    while vec:
        if vec & 1:
            summary ^= mat[mat_index]

        vec >>= 1
        mat_index += 1

    return summary


def _combine64(poly, initCrc, rev, xorOut, crc1, crc2, len2):
    if len2 == 0:
        return crc1

    even = [0] * GF2_DIM
    odd = [0] * GF2_DIM

    crc1 ^= initCrc ^ xorOut

    if (rev):
        # put operator for one zero bit in odd
        odd[0] = poly  # CRC-64 polynomial
        row = 1
        for n in xrange(1, GF2_DIM):
            odd[n] = row
            row <<= 1
    else:
        row = 2
        for n in xrange(0, GF2_DIM - 1):
            odd[n] = row
            row <<= 1
        odd[GF2_DIM - 1] = poly

    gf2_matrix_square(even, odd)

    gf2_matrix_square(odd, even)

    while True:
        gf2_matrix_square(even, odd)
        if len2 & long(1):
            crc1 = gf2_matrix_times(even, crc1)
        len2 >>= 1
        if len2 == 0:
            break

        gf2_matrix_square(odd, even)
        if len2 & long(1):
            crc1 = gf2_matrix_times(odd, crc1)
        len2 >>= 1

        if len2 == 0:
            break

    crc1 ^= crc2

    return crc1


def _verifyPoly(poly):
    msg = 'The degree of the polynomial must be 8, 16, 24, 32 or 64'
    poly = long(poly)  # Use a common representation for all operations
    for n in (8, 16, 24, 32, 64):
        low = long(1) << n
        high = low * 2
        if low <= poly < high:
            return n
    raise ValueError(msg)


def _bitrev(x, n):
    x = long(x)
    y = long(0)
    for i in xrange(n):
        y = (y << 1) | (x & long(1))
        x = x >> 1
    if ((long(1) << n) - 1) <= sys.maxint:
        return int(y)
    return y


def _verifyParams(poly, initCrc, xorOut):
    sizeBits = _verifyPoly(poly)

    mask = (long(1) << sizeBits) - 1

    initCrc = long(initCrc) & mask
    if mask <= sys.maxint:
        initCrc = int(initCrc)

    xorOut = long(xorOut) & mask
    if mask <= sys.maxint:
        xorOut = int(xorOut)

    return (sizeBits, initCrc, xorOut)


def gen_key(host, port):
    return '{}:{}'.format(host, port)


class CacheEntry(object):
    def __init__(self, host, port, ip_list, expire):
        self.host = host
        self.ip_list = ip_list
        self.port = port
        self.expire = expire
        self.immortal = False

    def remove(self, entry):
        if entry in self.ip_list and len(self.ip_list) > 1:
            self.ip_list.remove(entry)

    def copy_ip_list(self):
        return self.ip_list.copy()

    def get_key(self):
        return gen_key(self.host, self.port)


class DnsCacheService(object):
    def __init__(self):
        self.cache = {}
        self.async_started = False
        self.is_shutdown = True
        self.async_lock = threading.Lock()

    def shutdown(self):
        with self.async_lock:
            self.async_started = False
            self.is_shutdown = True

    def async_refresh_cache(self, dns_cache_time, interval=30):
        if self.async_started:
            return False

        with self.async_lock:
            if self.async_started:
                return False

            def _refresh_cache():
                while not self.is_shutdown:
                    try:
                        time.sleep(interval)
                        for k, v in self.cache.copy().items():
                            ip_list = resolve_ip_list(v.host, v.port)
                            if ip_list and len(ip_list) > 0:
                                self.cache[k] = CacheEntry(v.host, v.port, ip_list, time.time() + dns_cache_time)
                            else:
                                v.immortal = True
                    except Exception as ex:
                        logger.info('_refresh_cache exception, {}'.format(ex))

            self.async_started = True
            self.is_shutdown = False
            t = threading.Thread(target=_refresh_cache)
            t.setDaemon(True)
            t.start()
            return True

    def get_ip_list(self, host: str, port: int) -> CacheEntry:
        now = int(time.time())
        key = gen_key(host, port)
        if key in self.cache:
            info = self.cache[key]
            if info.immortal or info.expire >= now:
                return info
            self.cache.pop(key)

    def add(self, host, port, ip_list, expire):
        key = gen_key(host, port)
        if key in self.cache:
            return self.cache[key]
        entry = CacheEntry(host, port, ip_list, expire)
        self.cache[key] = entry
        logger.info('in-request: add cache address:{}'.format(key))
        return entry


def resolve_ip_list(host, port):
    try:
        family = allowed_gai_family()
        return socket.getaddrinfo(host, port, family, socket.SOCK_STREAM)
    except:
        return None


def check_enum_type(acl=None, storage_class=None, metadata_directive=None, az_redundancy=None,
                    permission=None, grantee=None, canned=None):
    if acl:
        check_acl_type(acl)

    if storage_class:
        check_storage_class_type(storage_class)

    if metadata_directive:
        check_metadata_directive_type(metadata_directive)

    if az_redundancy:
        check_az_redundancy_type(az_redundancy)

    if permission:
        check_permission_type(permission)

    if grantee:
        check_grantee_type(grantee)

    if canned:
        check_canned_type(canned)


def check_acl_type(obj):
    if not isinstance(obj, ACLType):
        raise TosClientError('invalid acl type')


def check_storage_class_type(obj):
    if not isinstance(obj, StorageClassType):
        raise TosClientError('invalid storage class')


def check_metadata_directive_type(obj):
    if not isinstance(obj, MetadataDirectiveType):
        raise TosClientError('invalid metadata directive type')


def check_az_redundancy_type(obj):
    if not isinstance(obj, AzRedundancyType):
        raise TosClientError('invalid az redundancy type')


def check_permission_type(obj):
    if not isinstance(obj, PermissionType):
        raise TosClientError('invalid permission type')


def check_grantee_type(obj):
    if not isinstance(obj, GranteeType):
        raise TosClientError('invalid grantee type')


def check_canned_type(obj):
    if not isinstance(obj, CannedType):
        raise TosClientError('invalid canned type')


def check_part_size(part_size):
    if not (part_size is not None and MIN_PART_SIZE <= part_size <= MAX_PART_SIZE):
        raise TosClientError('invalid part size, the size must be [5242880, 5368709120], size={}'.format(part_size))


def check_part_number(size, part_size):
    number = get_number(size, part_size)
    if number > MAX_PART_NUMBER:
        raise TosClientError('unsupported part number, the maximum is 10000')


def check_client_encryption_algorithm(algorithm):
    if algorithm:
        if algorithm not in CLIENT_ENCRYPTION_ALGORITHM:
            return TosClientError('invalid encryption-decryption algorithm')


def check_server_encryption_algorithm(algorithm):
    if algorithm:
        if algorithm not in SERVER_ENCRYPTION_ALGORITHM:
            return TosClientError('invalid encryption-decryption algorithm')


def is_ip(host):
    p = re.compile('^((25[0-5]|2[0-4]\d|[01]?\d\d?)\.){3}(25[0-5]|2[0-4]\d|[01]?\d\d?)$')
    if p.match(host):
        return True
    else:
        return False


class LogInfo(object):
    def __init__(self):
        self.start = time.perf_counter()

    def fail(self, func_name, e):
        end = time.perf_counter()
        logger.info('after-request: {}  exception: {}, usedTime:{} s'.format(func_name, e, end - self.start))
        raise e

    def success(self, func_name, res):
        end = time.perf_counter()
        logger.info(
            'after-request: {} exec httpCode: {}, requestId: {}, usedTime: {} s'.format(func_name, res.status,
                                                                                        res.request_id,
                                                                                        end - self.start))


class MergeProcess(object):
    def __init__(self, process, total_bytes, task_num, consumed_bytes):
        self.process = process
        self.totalBytes = total_bytes
        self.consumed_bytes = consumed_bytes
        if consumed_bytes == 0:
            self.status = DataTransferType.Data_Transfer_Init
        else:
            self.status = DataTransferType.Data_Transfer_RW
        self.taskNum = task_num
        self.count = 0
        self.lock = threading.Lock()

    def __call__(self, consumed_bytes, total_bytes, rw_once_bytes, type: DataTransferType):
        with self.lock:
            if type == DataTransferType.Data_Transfer_Started and self.status == DataTransferType.Data_Transfer_Init:
                self.status = DataTransferType.Data_Transfer_Started
                self.process(0, self.totalBytes, 0, self.status)
            elif type == DataTransferType.Data_Transfer_RW:
                self.status = DataTransferType.Data_Transfer_RW
                self.consumed_bytes += rw_once_bytes
                self.process(self.consumed_bytes, self.totalBytes, rw_once_bytes, DataTransferType.Data_Transfer_RW)

            elif type == DataTransferType.Data_Transfer_Succeed:
                self.count += 1
                if self.count == self.taskNum:
                    self.process(self.consumed_bytes, self.totalBytes, 0, DataTransferType.Data_Transfer_Succeed)


class UploadEventHandler(object):
    def __init__(self, cal_back_fuc, bucket, key, file_path, checkpoint_file, upload_id=None):
        super(UploadEventHandler, self).__init__()
        self.cal_back_fuc = cal_back_fuc
        self.bucket = bucket
        self.key = key
        self.upload_id = upload_id
        self.file_path = file_path
        self.checkpoint_file = checkpoint_file

    def _cal_back_event(self, event_type, e=None, part_info=None):
        self.cal_back_fuc(event_type, e, self.bucket, self.key, self.upload_id, self.file_path,
                          self.checkpoint_file, part_info)

    def __call__(self, *args, **kwargs):
        if self.cal_back_fuc:
            self._cal_back_event(*args, **kwargs)


class DownloadEventHandler(object):
    def __init__(self, cal_back_fuc, bucket, key, version_id, file_path, checkpoint_file):
        super(DownloadEventHandler, self).__init__()
        self.cal_back_fuc = cal_back_fuc
        self.bucket = bucket
        self.key = key
        self.version_id = version_id
        self.file_path = file_path
        self.checkpoint_file = checkpoint_file
        self.temp_file_path = file_path + '.temp'

    def _cal_back_event(self, event_type, e=None, part_info=None):
        self.cal_back_fuc(event_type, e, self.bucket, self.key, self.version_id,
                          self.file_path, self.checkpoint_file, self.temp_file_path, part_info)

    def __call__(self, *args, **kwargs):
        if self.cal_back_fuc:
            self._cal_back_event(*args, **kwargs)


class ResumableCopyObject(object):
    def __init__(self, cal_back_fuc, bucket, key, src_bucket, src_key, src_version_id, checkpoint_file, upload_id=None):
        super(ResumableCopyObject, self).__init__()
        self.cal_back_fuc = cal_back_fuc
        self.bucket = bucket
        self.key = key
        self.upload_id = upload_id
        self.src_bucket = src_bucket
        self.src_key = src_key
        self.src_version_id = src_version_id
        self.checkpoint_file = checkpoint_file

    def _cal_back_event(self, event_type, e=None, part_info=None):
        self.cal_back_fuc(event_type, e, self.bucket, self.key, self.upload_id,
                          self.src_bucket, self.src_key, self.src_version_id, self.checkpoint_file, part_info)

    def __call__(self, *args, **kwargs):
        if self.cal_back_fuc:
            self._cal_back_event(*args, **kwargs)


def _param_to_quoted_query(k, v):
    if v:
        return quote(str(k), '') + '=' + quote(str(v), '')
    else:
        return quote(k, '/~')


def _if_map(region: str, endpoint: str):
    if region in REGION_MAP and not endpoint:
        return REGION_MAP[region]
    else:
        return endpoint


def _format_endpoint(endpoint):
    if not endpoint.startswith('http://') and not endpoint.startswith('https://'):
        return 'https://' + endpoint
    else:
        return endpoint


def _make_uri(bucket=None, key=None):
    if bucket and not key:
        return '/{0}'.format(bucket)
    if bucket and key:
        return '/{0}/{1}'.format(bucket, key)
    return '/'


def _make_virtual_host_uri(key=None):
    if key:
        return '/{0}'.format(key)
    return '/'


def _get_host(endpoint):
    if endpoint.startswith('http://'):
        return endpoint[7:]
    if endpoint.startswith('https://'):
        return endpoint[8:]
    return endpoint


def _get_scheme(endpoint):
    if endpoint.startswith('http://'):
        return 'http://'
    if endpoint.startswith('https://'):
        return 'https://'
    return 'https://'


def _get_virtual_host(bucket, endpoint):
    if bucket:
        return bucket + '.' + _get_host(endpoint)
    else:
        return _get_host(endpoint)


def _cal_content_sha256(data):
    if data and hasattr(data, 'seek'):
        position = data.tell()
        read_chunksize = functools.partial(data.read,
                                           PAYLOAD_BUFFER)
        checksum = sha256()
        for chunk in iter(read_chunksize, b''):
            checksum.update(chunk)
        hex_checksum = checksum.hexdigest()
        data.seek(position)
        return hex_checksum
    elif data:
        return sha256(data).hexdigest()
    else:
        return EMPTY_SHA256_HASH


def _make_virtual_host_url(host, scheme, bucket=None, key=None):
    url = host
    if bucket and key:
        url = '{0}.{1}/{2}'.format(bucket, host, quote(key, '/~'))
    elif bucket and not key:
        url = '{0}.{1}'.format(bucket, host)
    elif key:
        url = '{0}/{1}'.format(host, quote(key, '/~'))

    return _format_endpoint(scheme + url)


def is_s3_endpoint(endpoint):
    if endpoint in S3_REGION_LIST:
        return True
    return False



# -*- coding: utf-8 -*-

import logging
import os
import unittest

import tos
from tests.common import TosTestBase, random_string, random_bytes

tos.set_logger(level=logging.DEBUG)


class TestObjectCrc(TosTestBase):
    def test_getobject_crc(self):
        client = tos.TosClientV2(self.ak, self.sk, self.endpoint, self.region, enable_crc=True)
        bucket = self.bucket_name + '-test-crc'
        client.create_bucket(bucket)
        self.bucket_delete.append(bucket)
        content = random_bytes(1024 * 1024 * 6)
        file_name = self.random_filename()
        with open(file_name, 'wb') as f:
            f.write(content)
        for i in range(2):
            key = self.random_key('.js')
            client.put_object(bucket, key, content=content)
            out = client.get_object(bucket, key)
            # 整个读取
            self.assertEqual(out.read(), content)
            out = client.get_object(bucket, key)

            # 不使用迭代器， chuck读取
            con = b''
            while True:
                buf = out.read(8 * 1024)
                if not buf:
                    break
                con += buf
            self.assertEqual(con, content)

            # 使用迭代器读取
            out = client.get_object(bucket, key)
            con = b''
            for buf in out:
                con += buf
            self.assertEqual(con, content)

            # 查询部分数据
            out = client.get_object(bucket, key, range_start=100, range_end=300)
            self.assertEqual(out.read(), content[100:301])

            out = client.get_object(bucket, key, range_start=100, range_end=300)
            con = b''
            while True:
                buf = out.read(8 * 1024)
                if not buf:
                    break
                con += buf
            self.assertEqual(con, content[100:301])

            out = client.get_object(bucket, key, range_start=100, range_end=300)
            con = b''
            for buf in out:
                con += buf
            self.assertEqual(con, content[100:301])

            client.put_object_from_file(bucket, key, file_name)
            cwd = os.getcwd()
            client.get_object_to_file(bucket, key, cwd)
            self.assertFileContent(cwd + '/' + key, content)
            client = tos.TosClientV2(self.ak, self.sk, self.endpoint, self.region, enable_crc=False)

    def random_key(self, suffix=''):
        key = self.prefix + random_string(12) + suffix
        return key


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

import datetime
import json
import unittest
from unittest import mock

import pytz
from tos.exceptions import TosError

from tests.common import MockResponse, TosTestCase


class TestObject(TosTestCase):
    @mock.patch('requests.Session.request')
    def test_put_object(self, mock_request):
        mock_request.return_value = MockResponse(headers={
            'ETag': '"etag"',
            'x-tos-server-side-encryption-customer-algorithm': 'AES256',
            'x-tos-server-side-encryption-customer-key': 'key',
            'x-tos-server-side-encryption-customer-key-md5': 'eb5875f724a1a982'
        })

        res = self.client.put_object(Bucket=self.bucket_name, Key=self.key_name, Body=b'aaa.txt')
        self.assertEqual(res.etag, 'etag')
        self.assertEqual(res.sse_customer_algorithm, 'AES256')
        self.assertEqual(res.sse_customer_key, 'key')
        self.assertEqual(res.sse_customer_key_md5, 'eb5875f724a1a982')

    @mock.patch('requests.Session.request')
    def test_get_object(self, mock_request):
        mock_request.return_value = MockResponse(headers={'content-range': 'bytes=0-100'})

        res = self.client.get_object(Bucket=self.bucket_name, Key=self.key_name)
        self.assertEqual(res.status, 200)
        self.assertEqual(res.content_range, 'bytes=0-100')

    @mock.patch('requests.Session.request')
    def test_get_object_not_found(self, mock_request):
        mock_request.return_value = MockResponse(status_code=404)

        try:
            self.client.get_object(Bucket=self.bucket_name, Key=self.key_name)
            self.assertTrue(False)
        except TosError as e:
            self.assertEqual(e.status, 404)

    @mock.patch('requests.Session.request')
    def test_head_object(self, mock_request):
        mock_request.return_value = MockResponse(headers={
            'ETag': '"etag"',
            'content-type': 'text/html',
            'content-length': '100',
            'x-tos-server-side-encryption-customer-algorithm': 'AES256',
            'x-tos-server-side-encryption-customer-key': 'key',
            'x-tos-server-side-encryption-customer-key-md5': 'eb5875f724a1a982',
            'x-tos-meta-key': 'self-value',
            'last-modified': 'Fri, 01 Jan 2021 00:00:00 GMT',
            'expires': 'Fri, 01 Jan 2021 00:00:00 GMT'
        })

        res = self.client.head_object(Bucket=self.bucket_name, Key=self.key_name)
        self.assertEqual(res.content_type, 'text/html')
        self.assertEqual(res.content_length, 100)
        self.assertEqual(res.etag, 'etag')
        self.assertEqual(res.sse_customer_algorithm, 'AES256')
        self.assertEqual(res.sse_customer_key, 'key')
        self.assertEqual(res.sse_customer_key_md5, 'eb5875f724a1a982')

        self.assertEqual(res.metadata['key'], 'self-value')
        self.assertEqual(res.last_modified, datetime.datetime(2021, 1, 1, tzinfo=pytz.utc))
        self.assertEqual(res.expires, datetime.datetime(2021, 1, 1, tzinfo=pytz.utc))

    @mock.patch('requests.Session.request')
    def test_delete_object(self, mock_request):
        mock_request.return_value = MockResponse(status_code=204)

        res = self.client.delete_object(Bucket=self.bucket_name, Key=self.key_name, VersionId='df78ca5')
        self.assertEqual(res.status, 204)

    @mock.patch('requests.Session.request')
    def test_copy_object(self, mock_request):
        body = {
            'ETag': '"etag"',
            'LastModified': '2021-01-01T00:00:00.000Z',
        }
        mock_request.return_value = MockResponse(body=json.dumps(body))

        copySource = {
            'Bucket': 'src-bkt',
            'Key': 'src-key',
            'VersionId': 'src_id'
        }
        res = self.client.copy_object(Bucket=self.bucket_name, CopySource=copySource, Key=self.key_name)
        self.assertEqual(res.etag, 'etag')
        self.assertEqual(res.last_modified, datetime.datetime(2021, 1, 1, tzinfo=pytz.utc))

    @mock.patch('requests.Session.request')
    def test_append_object(self, mock_request):
        mock_request.return_value = MockResponse(headers={
            'ETag': '"etag"',
            'x-tos-server-side-encryption-customer-algorithm': 'AES256',
            'x-tos-server-side-encryption-customer-key': 'key',
            'x-tos-server-side-encryption-customer-key-md5': 'eb5875f724a1a982'
        })

        res = self.client.append_object(Bucket=self.bucket_name, Key=self.key_name, Offset=100, Body=b'aaa')
        self.assertEqual(res.etag, 'etag')
        self.assertEqual(res.sse_customer_algorithm, 'AES256')
        self.assertEqual(res.sse_customer_key, 'key')
        self.assertEqual(res.sse_customer_key_md5, 'eb5875f724a1a982')

    @mock.patch('requests.Session.request')
    def test_put_object_acl(self, mock_request):
        mock_request.return_value = MockResponse()

        res = self.client.put_object_acl(self.bucket_name, self.key_name, ACL='public-read', VersionId='df78ca5')
        self.assertEqual(res.status, 200)

    @mock.patch('requests.Session.request')
    def test_get_object_acl(self, mock_request):
        acl = {
            'Grants': [
                {
                    'Grantee': {
                        'ID': 'acl id',
                        'DisplayName': 'acl name',
                        'Type': 'Group',
                        'Canned': 'AllUsers'
                    },
                    'Permission': 'FULL_CONTROL'
                },
            ],
            'Owner': {
                'ID': 'id',
                'DisplayName': 'name'
            }
        }
        mock_request.return_value = MockResponse(body=json.dumps(acl))

        res = self.client.get_object_acl(self.bucket_name, self.key_name, VersionId='df78ca5')
        self.assertEqual(res.owner.id, 'id')
        self.assertEqual(res.owner.name, 'name')
        self.assertEqual(len(res.grant_list), 1)
        self.assertEqual(res.grant_list[0].permission, 'FULL_CONTROL')
        self.assertEqual(res.grant_list[0].grantee.id, 'acl id')
        self.assertEqual(res.grant_list[0].grantee.display_name, 'acl name')
        self.assertEqual(res.grant_list[0].grantee.type, 'Group')
        self.assertEqual(res.grant_list[0].grantee.canned, 'AllUsers')

    @mock.patch('requests.Session.request')
    def test_delete_objects(self, mock_request):
        mock_data = {
            'Deleted': [{
                'Key': 'key1',
                'VersionId': 'versionId1',
            }],
            'Error': [{
                'Code': 'NoSuchKey',
                'Message': 'No such key',
                'Key': 'key2',
                'VersionId': 'versionId2'
            }]
        }
        mock_request.return_value = MockResponse(body=json.dumps(mock_data))

        delete = {
            'Objects': [
                {
                    'Key': 'key1',
                    'VersionId': 'versionId1'
                },
                {
                    'Key': 'key2',
                    'VersionId': 'versionId2'
                }
            ],
            'Quiet': False
        }
        res = self.client.delete_objects(Bucket=self.bucket_name, Delete=delete)
        self.assertEqual(len(res.deleted_list), 1)
        self.assertEqual(res.deleted_list[0].key, 'key1')
        self.assertEqual(res.deleted_list[0].version_id, 'versionId1')
        self.assertEqual(len(res.error_list), 1)
        self.assertEqual(res.error_list[0].code, 'NoSuchKey')
        self.assertEqual(res.error_list[0].message, 'No such key')
        self.assertEqual(res.error_list[0].key, 'key2')
        self.assertEqual(res.error_list[0].version_id, 'versionId2')


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

import datetime
import json
import unittest
from unittest import mock

import pytz

from tests.common import MockResponse, TosTestCase


class TestMultipart(TosTestCase):
    @mock.patch('requests.Session.request')
    def test_create_multipart_upload(self, mock_request):
        mock_data = {
            'UploadId': 'uploadId',
            'Bucket': self.bucket_name,
            'Key': self.key_name,
        }
        mock_request.return_value = MockResponse(body=json.dumps(mock_data))

        res = self.client.create_multipart_upload(Bucket=self.bucket_name, Key=self.key_name)
        self.assertEqual(res.upload_id, 'uploadId')
        self.assertEqual(res.bucket, self.bucket_name)
        self.assertEqual(res.key, self.key_name)

    @mock.patch('requests.Session.request')
    def test_upload_part(self, mock_request):
        mock_request.return_value = MockResponse(headers={
            'ETag': '"etag"',
            'x-tos-server-side-encryption-customer-algorithm': 'AES256',
            'x-tos-server-side-encryption-customer-key': 'key',
            'x-tos-server-side-encryption-customer-key-md5': 'md5',
        })

        res = self.client.upload_part(Bucket=self.bucket_name, Key=self.key_name, PartNumber=1, UploadId='uploadId',
                                      Body=b'data')
        self.assertEqual(res.etag, 'etag')
        self.assertEqual(res.sse_customer_algorithm, 'AES256')
        self.assertEqual(res.sse_customer_key, 'key')
        self.assertEqual(res.sse_customer_key_md5, 'md5')

    @mock.patch('requests.Session.request')
    def test_upload_part_copy(self, mock_request):
        mock_data = {
            'ETag': '\"etag\"',
            'LastModified': '2021-01-01T00:00:00.000Z',
        }
        mock_request.return_value = MockResponse(body=json.dumps(mock_data))

        copy_source = {
            'Bucket': 'src-bucket',
            'Key': 'src-key',
            'VersionId': 'version-id'
        }
        res = self.client.upload_part_copy(Bucket=self.bucket_name, CopySource=copy_source, Key=self.key_name,
                                           PartNumber=1,
                                           UploadId='uploadId')
        self.assertEqual(res.etag, 'etag')
        self.assertEqual(res.last_modified, datetime.datetime(2021, 1, 1, tzinfo=pytz.utc))

    @mock.patch('requests.Session.request')
    def test_complete_multipart_upload(self, mock_request):
        mock_data = {
            'Location': self.location,
            'Bucket': self.bucket_name,
            'Key': self.key_name,
            'ETag': '"etag"'
        }
        mock_request.return_value = MockResponse(body=json.dumps(mock_data))

        upload = {
            'Parts': [
                {
                    'ETag': 'etag-1',
                    'PartNumber': 1
                },
                {
                    'ETag': 'etag-2',
                    'PartNumber': 2
                },
            ]
        }
        res = self.client.complete_multipart_upload(Bucket=self.bucket_name, Key=self.key_name, UploadId='upload-id',
                                                    MultipartUpload=upload)
        self.assertEqual(res.location, self.location)
        self.assertEqual(res.bucket, self.bucket_name)
        self.assertEqual(res.key, self.key_name)
        self.assertEqual(res.etag, 'etag')

    @mock.patch('requests.Session.request')
    def test_list_multipart_uploads(self, mock_request):
        upload_id = 'upload-id'
        delimiter = '/'
        encoding_type = 'url'
        key_marker = 'key-marker'
        next_key_marker = 'next-key-marker'
        max_uploads = 10
        prefix = 'abc'
        upload_id_marker = 'upload-marker'
        next_upload_id_marker = 'next-upload-marker'
        storage_class = 'STANDARD'
        mock_data = {
            'Bucket': self.bucket_name,
            'UploadIdMarker': upload_id_marker,
            'NextKeyMarker': next_key_marker,
            'NextUploadIdMarker': next_upload_id_marker,
            'Delimiter': delimiter,
            'Prefix': prefix,
            'MaxUploads': max_uploads,
            'IsTruncated': True,
            'CommonPrefixes': [{
                'Prefix': prefix
            }],
            'Uploads': [{
                'Key': self.key_name,
                'UploadId': upload_id,
                'StorageClass': storage_class,
                'Initiated': '2021-01-01T00:00:00.000Z',
                'Owner': {
                    'ID': 'id',
                    'DisplayName': 'name'
                },

            }]
        }
        mock_request.return_value = MockResponse(body=json.dumps(mock_data))

        res = self.client.list_multipart_uploads(Bucket=self.bucket_name, Delimiter=delimiter,
                                                 EncodingType=encoding_type, KeyMarker=key_marker,
                                                 MaxUploads=max_uploads, Prefix=prefix, UploadIdMarker=upload_id_marker)
        self.assertEqual(res.bucket, self.bucket_name)
        self.assertEqual(res.upload_id_marker, upload_id_marker)
        self.assertEqual(res.next_key_marker, next_key_marker)
        self.assertEqual(res.next_upload_id_marker, next_upload_id_marker)
        self.assertEqual(res.delimiter, delimiter)
        self.assertEqual(res.prefix, prefix)
        self.assertEqual(res.max_uploads, max_uploads)
        self.assertEqual(res.is_truncated, True)

        self.assertEqual(len(res.upload_list), 1)
        self.assertEqual(res.upload_list[0].key, self.key_name)
        self.assertEqual(res.upload_list[0].initiated, datetime.datetime(2021, 1, 1, tzinfo=pytz.utc))
        self.assertEqual(res.upload_list[0].storage_class, storage_class)
        self.assertEqual(res.upload_list[0].upload_id, upload_id)
        self.assertEqual(res.upload_list[0].owner.id, 'id')
        self.assertEqual(res.upload_list[0].owner.name, 'name')

        self.assertEqual(len(res.common_prefix_list), 1)
        self.assertEqual(res.common_prefix_list[0].prefix, prefix)

    @mock.patch('requests.Session.request')
    def test_list_parts(self, mock_request):
        upload_id = 'upload-id'
        max_parts = 10
        part_number_marker = 11
        next_part_number_marker = 21
        storage_class = 'STANDARD'
        mock_data = {
            'Bucket': self.bucket_name,
            'Key': self.key_name,
            'UploadId': upload_id,
            'PartNumberMarker': part_number_marker,
            'NextPartNumberMarker': next_part_number_marker,
            'MaxParts': max_parts,
            'StorageClass': storage_class,
            'IsTruncated': True,
            'Owner': {
                'ID': 'id',
                'DisplayName': 'name'
            },
            'Parts': [
                {
                    'PartNumber': 11,
                    'LastModified': '2021-01-01T00:00:00.000Z',
                    'ETag': '"etag-1"',
                    'Size': 100
                }
            ]
        }
        mock_request.return_value = MockResponse(body=json.dumps(mock_data))

        res = self.client.list_parts(Bucket=self.bucket_name, Key=self.key_name, UploadId=upload_id, MaxParts=max_parts,
                                     PartNumberMarker=part_number_marker)
        self.assertEqual(res.bucket, self.bucket_name)
        self.assertEqual(res.key, self.key_name)
        self.assertEqual(res.upload_id, upload_id)
        self.assertEqual(res.part_number_marker, part_number_marker)
        self.assertEqual(res.next_part_number_marker, next_part_number_marker)
        self.assertEqual(res.max_parts, max_parts)
        self.assertEqual(res.storage_class, storage_class)
        self.assertEqual(res.is_truncated, True)

        self.assertEqual(res.owner.id, 'id')
        self.assertEqual(res.owner.name, 'name')

        self.assertEqual(len(res.part_list), 1)
        self.assertEqual(res.part_list[0].part_number, 11)
        self.assertEqual(res.part_list[0].last_modified, datetime.datetime(2021, 1, 1, tzinfo=pytz.utc))
        self.assertEqual(res.part_list[0].etag, 'etag-1')
        self.assertEqual(res.part_list[0].size, 100)

    @mock.patch('requests.Session.request')
    def test_abort_multipart_upload(self, mock_request):
        mock_request.return_value = MockResponse(status_code=204)

        res = self.client.abort_multipart_upload(Bucket=self.bucket_name, Key=self.key_name, UploadId='upload-id')
        self.assertEqual(res.status, 204)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import base64
import datetime
import http
import http.client as httplib
import os
import time
import unittest
from io import StringIO, BytesIO

import requests

import tos
from tests.common import TosTestBase, random_string, random_bytes, calculate_md5
from tos import TosClientV2
from tos.consts import MIN_TRAFFIC_LIMIT
from tos.credential import EnvCredentialsProvider
from tos.enum import (ACLType, AzRedundancyType, DataTransferType,
                      GranteeType, MetadataDirectiveType, PermissionType,
                      StorageClassType, VersioningStatusType, TierType, CopyEventType, HttpMethodType)
from tos.exceptions import TosClientError, TosServerError
from tos.models2 import Deleted, Grant, Grantee, ListObjectsOutput, Owner, ObjectTobeDeleted, Tag, \
    PostSignatureCondition, UploadedPart, PolicySignatureCondition, RestoreJobParameters
from tos.utils import RateLimiter


def get_socket_io():
    conn = http.client.HTTPConnection('tos-cn-beijing.volces.com', 80)
    conn.request('GET', '/')
    content = conn.getresponse()
    return content


def _get_host_schema(endpoint):
    if endpoint.startswith('http://'):
        return 'http://', endpoint[7:]
    if endpoint.startswith('https://'):
        return 'https://', endpoint[8:]
    return 'http://', endpoint


class TestObject(TosTestBase):
    def test_object(self):
        bucket_name = self.bucket_name + '-test-object'
        self.bucket_delete.append(bucket_name)
        key = self.random_key('.js')
        content = random_bytes(1024)

        self.client.create_bucket(bucket_name)
        put_object_out = self.client.put_object(bucket_name, key=key, content=content)
        self.assertTrue(len(put_object_out.etag) > 0)
        self.assertTrue(len(put_object_out.id2) > 0)
        self.assertTrue(put_object_out.hash_crc64_ecma > 0)

        head_object_out = self.client.head_object(bucket_name, key)
        self.assertTrue(len(head_object_out.etag) > 0)
        self.assertTrue(head_object_out.hash_crc64_ecma > 0)
        self.assertTrue(len(head_object_out.meta) == 0)
        self.assertTrue(len(head_object_out.object_type) > 0)
        self.assertTrue(head_object_out.last_modified is not None)
        self.assertTrue(head_object_out.object_type == 'Normal')
        self.assertTrue(head_object_out.delete_marker is False)
        self.assertTrue(head_object_out.content_length == 1024)

        get_object_out = self.client.get_object(bucket_name, key)
        res = b''
        for chuck in get_object_out:
            res = res + chuck

        self.assertEqual(res, content)

        range_out = self.client.get_object(bucket_name, key, range_start=1, range_end=100)
        read_content = range_out.read()
        self.assertEqual(read_content, content[1:101])

    def test_put_with_meta(self):
        bucket_name = self.bucket_name + '-put-object-with-meta'
        key = "张三.txt"
        content = random_string(123)
        conn = httplib.HTTPConnection('tos-cn-beijing.volces.com', 80)
        conn.request('GET', '/')
        content_io = conn.getresponse()
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        for i in range(2):
            raw = "!@#$%^&*()_+-=[]{}|;':\",./<>?中文测试编码%20%%%^&abcd /\\"
            meta = {'name': ' %张/三%', 'age': '12', 'special': raw, raw: raw}
            self.client.put_object(bucket_name, key=key,
                                   content=content,
                                   meta=meta
                                   )
            get_object_out = self.client.get_object(bucket_name, key)
            m = get_object_out.meta
            self.assertEqual(m['name'], meta['name'])
            self.assertEqual(m['age'], meta['age'])
            self.assertEqual(m['special'], meta['special'])
            self.assertEqual(m[raw], meta[raw])

            meta['name'] = '李四'
            self.client.set_object_meta(bucket_name, key, meta=meta)
            head_out = self.client.head_object(bucket_name, key)
            m = head_out.meta
            self.assertEqual(m['name'], meta['name'])
            self.assertEqual(m['age'], meta['age'])
            self.assertEqual(m['special'], meta['special'])
            self.assertEqual(m[raw], meta[raw])
            key = '李四.txt'
            content = content_io

    def test_with_string_io(self):
        io = StringIO('a')
        io.seek(0)
        bucket_name = self.bucket_name + 'string-io'
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        self.client.put_object(bucket=bucket_name, key="2", content=io)
        out = self.client.get_object(bucket=bucket_name, key='2')
        self.assertEqual(out.read(), b'a')
        self.client.put_object(bucket=bucket_name, key='4', content=b'')

    def test_put_with_options(self):
        bucket_name = self.bucket_name + '-put-with-options'
        key = self.random_key()
        content = random_bytes(123)

        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        self.client.put_object(bucket_name, key=key,
                               content=content,
                               cache_control='CacheControl',
                               acl=ACLType.ACL_Private,
                               content_encoding='utf-8',
                               content_disposition='attachment; filename=张123.txt',
                               content_length=123,
                               content_language='english',
                               expires=datetime.datetime(2023, 1, 29),
                               website_redirect_location='/test',
                               )
        get_object_out = self.client.get_object(bucket_name, key)
        self.assertTrue(len(get_object_out.cache_control) > 0)
        self.assertEqual(get_object_out.content_encoding, 'utf-8')
        self.assertEqual(get_object_out.content_disposition, 'attachment; filename=张123.txt')
        self.assertEqual(get_object_out.content_language, 'english')
        self.assertEqual(get_object_out.website_redirect_location, '/test')

    def test_put_with_cryptography(self):
        bucket_name = self.bucket_name + '-put-object-with-test-put-with-cryptography'
        key = self.random_key()
        content = random_bytes(100)

        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)

        with self.assertRaises(TosServerError):
            self.client.put_object(bucket_name, key=key, content=content, content_md5=random_bytes(20))

        with self.assertRaises(TosServerError) as cm:
            self.client.put_object(bucket_name, key=key, content=content, content_sha256=random_string(10))

        with self.assertRaises(TosServerError):
            self.client.put_object(bucket_name, key=key, content=content, ssec_algorithm="DEC")

    def test_put_with_empty_content(self):
        bucket_name = self.bucket_name + '-put-empty-object'
        key = self.random_key()
        conn = httplib.HTTPConnection('tos-cn-beijing.volces.com', 80)
        conn.request('GET', '/')
        content_io = conn.getresponse()
        content = b''

        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        for i in range(2):
            self.client.put_object(bucket_name, key=key, content=content)
            get_object_out = self.client.get_object(bucket_name, key)
            if self.client.enable_crc:
                self.assertEqual(get_object_out.client_crc, 0)
            self.assertEqual(get_object_out.hash_crc64_ecma, 0)
            content = content_io
            content.read()
            key = self.random_key()
        conn.close()
        file_name = self.random_filename()
        file = open(file_name, 'w')
        file.close()
        self.client.put_object_from_file(bucket_name, key, file_path=file_name)
        os.remove(file_name)

    def test_put_with_illegal_name(self):
        bucket_name = self.bucket_name + '-put-object-with-illegal-name'
        key = ''
        content = random_bytes(100)

        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)

        with self.assertRaises(TosClientError):
            self.client.put_object(bucket_name, key, content=content)

        key = '/+'
        self.client.put_object(bucket_name, key, content=content)

        key = '.'
        with self.assertRaises(TosClientError):
            self.client.put_object(bucket_name, key, content=content)

        with self.assertRaises(TosClientError):
            self.client.get_object(bucket_name, key)

        key = "中文测试"
        self.client.put_object(bucket_name, key, content=content)
        self.client.delete_object(bucket_name, key)

        key = '%1 ? # *'
        self.client.put_object(bucket_name, key, content=content)
        self.client.get_object(bucket_name, key)
        self.client.delete_object(bucket_name, key)

    def test_put_with_server_encryption(self):
        bucket_name = self.bucket_name + '-put-object-with-server-encryption'
        key = self.random_key('.js')
        content = random_bytes(100)
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)

        self.client.put_object(bucket_name, key, content=content, server_side_encryption="AES256")
        self.client.get_object(bucket_name, key)

        self.client.delete_object(bucket_name, key)

        key = self.random_key(".js")
        with self.assertRaises(TosServerError):
            self.client.put_object(bucket_name, key, content=content, ssec_algorithm="AES256")

    def test_put_with_data_transfer_listener(self):
        self.client.enable_crc = False
        bucket_name = self.bucket_name + '-put-object-with-transfer-listener'
        key = self.random_key('.js')
        key_2 = self.random_key('.js')
        content = random_bytes(1024 * 100)
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)

        def progress(consumed_bytes, total_bytes, rw_once_bytes,
                     data_type: DataTransferType):
            print("consumed_bytes:{0},total_bytes{1}, rw_once_bytes:{2}, type:{3}".format(consumed_bytes, total_bytes,
                                                                                          rw_once_bytes, data_type))

        self.client.put_object(bucket_name, key, content=content, data_transfer_listener=progress)

        out = self.client.get_object(bucket_name, key, data_transfer_listener=progress)

        read_info = bytes()
        for buf in out.content:
            read_info = read_info + buf
        self.assertEqual(read_info, content)

        out = self.client.get_object(bucket_name, key)
        self.client.put_object(bucket_name, key_2, content=out, data_transfer_listener=progress)
        get_out = self.client.get_object(bucket_name, key_2, data_transfer_listener=progress)
        read_info = bytes()
        for buf in get_out.content:
            read_info = read_info + buf
        self.assertEqual(read_info, content)

    def test_object_from_file(self):
        bucket_name = self.bucket_name + '-put-object-from-file'
        key = self.random_key('.js')
        file_name = self.random_filename()
        content = random_bytes(100)
        with open(file_name, 'wb') as fw:
            fw.write(content)

        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)

        limiter = RateLimiter(8 * 1024, 20 * 1024)
        self.client.put_object_from_file(bucket=bucket_name, key=key, file_path=file_name, rate_limiter=limiter)

        get_file_name = self.random_filename()
        self.client.get_object_to_file(bucket_name, key, get_file_name)
        self.assertFileContent(get_file_name, content)
        self.assertObjectContent(bucket_name, key, content)

    def test_object_from_file_with_large_file(self):
        bucket_name = self.bucket_name + '-put-object-from-file'
        key = self.random_key('.js')
        file_name = self.random_filename()
        content = random_bytes(1024 * 1024 * 10)
        with open(file_name, 'wb') as fw:
            fw.write(content)

        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        self.client.put_object_from_file(bucket=bucket_name, key=key, file_path=file_name)
        get_file_name = self.random_filename()
        self.client.get_object_to_file(bucket_name, key, get_file_name)
        object_out = self.client.get_object(bucket_name, key)
        while True:
            if not object_out.read(64 * 1024):
                break
        # assert object_out.client_crc == object_out.hash_crc64_ecma
        self.assertFileContent(get_file_name, content)
        self.assertObjectContent(bucket_name, key, content)

    def test_object_with_io(self):
        bucket_name = self.bucket_name + '-put-object-from-file'
        key = self.random_key('.js')
        file_name = self.random_filename()
        content = random_bytes(1024 * 1024)
        with open(file_name, 'wb') as fw:
            fw.write(content)
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        self.client.put_object_from_file(bucket=bucket_name, key=key, file_path=file_name)
        input = self.client.get_object(bucket_name, key)
        self.client.put_object(bucket_name, key + '1', content=input)
        self.assertObjectContent(bucket_name, key + '1', content)

    def test_with_stream(self):
        bucket_name = self.bucket_name + 'test-with-stream'
        key = self.random_key('.js')
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        conn = http.client.HTTPConnection('www.volcengine.com', 80)
        conn.request('GET', '/')
        content = conn.getresponse()

        def generator():
            var1 = "5\r\n"
            var2 = "x=FOO\r\n"
            var3 = "0\r\n\r\n"
            x = var1.encode('utf8')
            y = var2.encode('utf8')
            z = var3.encode('utf8')
            yield x
            yield y
            yield z

        self.client.close()
        self.client.put_object(bucket_name, key, content=generator())
        out = self.client.create_multipart_upload(bucket_name, key)
        self.client.upload_part(bucket_name, key, content=generator(), upload_id=out.upload_id, part_number=1)
        out = self.client.get_object(bucket_name, key)
        buf = b''
        for i in generator():
            buf += i
        self.assertEqual(out.read(), buf)

        self.client.put_object(bucket_name, key, content=content)
        out = self.client.get_object(bucket_name, key)
        conn = http.client.HTTPConnection('www.volcengine.com', 80)
        conn.request('GET', '/')
        content = conn.getresponse()
        buf = b''
        for chuck in content:
            buf += chuck
        self.assertEqual(len(buf), len(out.read()))

        input = requests.get('https://www.volcengine.com')
        self.client.put_object(bucket_name, key, content=input)
        out = self.client.get_object(bucket_name, key)
        self.assertEqual(len(out.read()), len(requests.get('https://www.volcengine.com').text))
        conn.close()

    def test_object_with_iterm(self):
        bucket_name = self.bucket_name + '-with-itern'
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)

        class Crc64IO(object):
            def __init__(self, data: BytesIO):
                self.data = data
                self.crc64 = tos.utils.Crc64()

            def __iter__(self):
                return self

            def __next__(self):
                return self.next()

            def next(self):
                content = self.data.read(1)
                if content:
                    self.crc64.update(content)
                    return content
                raise StopIteration

        io = BytesIO(b'123')
        io.seek(0)
        self.client.put_object(bucket=bucket_name, key='1234', content=Crc64IO(io))
        self.assertEqual(self.client.get_object(bucket_name, key='1234').read(), b'123')

    def test_delete_multi_objects(self):
        bucket_name = self.bucket_name + "-delete-multi-objects"
        key_1 = self.random_key('.js')
        key_2 = self.random_key('.js')
        content = random_bytes(20)

        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)

        self.client.put_object(bucket=bucket_name, key=key_1, content=content)
        self.client.put_object(bucket=bucket_name, key=key_2, content=content)
        self.client.head_object(bucket_name, key_1)
        object = []
        object.append(ObjectTobeDeleted(key_1))
        object.append(ObjectTobeDeleted(key_1))
        self.client.delete_multi_objects(bucket=bucket_name, objects=object, quiet=False)

    def test_copy_object(self):
        bucket_name_1 = self.bucket_name + '-copy-object1'
        bucket_name_2 = self.bucket_name + '-copy-object2'
        key = self.random_key(".java")
        content = random_bytes(1024 * 1024 * 5)
        self.client.create_bucket(bucket_name_1)
        self.bucket_delete.append(bucket_name_1)
        self.client.create_bucket(bucket_name_2)
        self.bucket_delete.append(bucket_name_2)
        self.client.put_object(bucket_name_1, key, content=content)
        self.client.copy_object(bucket_name_2, key, bucket_name_1, key)
        self.client.head_object(bucket_name_2, key)
        self.assertObjectContent(bucket_name_2, key, content)

        self.client.delete_object(bucket_name_1, key)

        with self.assertRaises(TosServerError):
            self.client.copy_object(bucket_name_2, key, bucket_name_1, key)

    def test_copy_object_options(self):
        bucket_name_1 = self.bucket_name + '-copy-object1'
        bucket_name_2 = self.bucket_name + '-copy-object2'
        key = self.random_key(".java")
        content = random_bytes(100)
        self.client.create_bucket(bucket_name_1)
        self.bucket_delete.append(bucket_name_1)
        self.client.create_bucket(bucket_name_2)
        self.bucket_delete.append(bucket_name_2)
        meta = {'姓名': '张三'}
        self.client.put_object(bucket_name_1, key=key,
                               content=content,
                               cache_control='CacheControl',
                               acl=ACLType.ACL_Private,
                               content_encoding='utf-8',
                               content_disposition='/sunyushantest',
                               content_length=100,
                               content_language='english',
                               expires=datetime.datetime(2023, 1, 29),
                               website_redirect_location='/test',
                               meta=meta
                               )

        out = self.client.copy_object(bucket_name_2, key, src_bucket=bucket_name_1, src_key=key,
                                      metadata_directive=MetadataDirectiveType.Metadata_Directive_Copy)
        self.client.head_object(bucket_name_2, key)
        get_object_out = self.client.get_object(bucket_name_2, key)

        self.assertTrue(len(get_object_out.cache_control) > 0)
        self.assertEqual(get_object_out.content_encoding, 'utf-8')
        self.assertEqual(get_object_out.content_disposition, '/sunyushantest')
        self.assertEqual(get_object_out.content_language, 'english')
        self.assertEqual(get_object_out.meta['姓名'], meta['姓名'])
        self.assertEqual(get_object_out.read(), content)

    def test_copy_with_set_option(self):
        bucket_name_1 = self.bucket_name + '-copy-object1'
        bucket_name_2 = self.bucket_name + '-copy-object2'
        key = self.random_key(".java")
        content = random_bytes(100)
        self.client.create_bucket(bucket_name_1)
        self.bucket_delete.append(bucket_name_1)
        self.client.create_bucket(bucket_name_2)
        self.bucket_delete.append(bucket_name_2)
        meta = {'姓名': '张三'}
        self.client.put_object(bucket_name_1, key=key,
                               content=content)

        self.client.copy_object(bucket_name_2, key, src_bucket=bucket_name_1, src_key=key,
                                cache_control='CacheControl',
                                acl=ACLType.ACL_Private,
                                content_encoding='utf-8',
                                content_disposition='/sunyushantest',
                                content_language='english',
                                expires=datetime.datetime(2023, 1, 29),
                                website_redirect_location='/test',
                                meta=meta,
                                storage_class=StorageClassType.Storage_Class_Ia,
                                metadata_directive=MetadataDirectiveType.Metadata_Directive_Replace
                                )

        self.client.head_object(bucket_name_2, key)
        get_object_out = self.client.get_object(bucket_name_2, key)

        self.assertTrue(len(get_object_out.cache_control) > 0)
        self.assertEqual(get_object_out.content_encoding, 'utf-8')
        self.assertEqual(get_object_out.content_disposition, '/sunyushantest')
        self.assertEqual(get_object_out.content_language, 'english')
        self.assertEqual(get_object_out.meta['姓名'], meta['姓名'])

        self.assertEqual(get_object_out.read(), content)

    # # 目前不知道如何开启多版本, 后续补充测试
    # def test_mult_version(self):
    #     bucket_name = self.bucket_name + '-test-mult-object'
    #     key = self.random_key('.js')
    #     content = random_bytes(1024)
    #     self.client.create_bucket(bucket_name)
    #     self.bucket_delete.append(bucket_name)
    #     out = self.version_client.put_bucket_versioning(bucket_name, enable=True)
    #
    #     time.sleep(30)
    #     put_object_out_v1 = self.client.put_object(bucket_name, key=key, content=content)
    #     version_1 = put_object_out_v1.version_id
    #     content = random_bytes(2048)
    #     put_object_out_v2 = self.client.put_object(bucket_name, key=key, content=content)
    #     version_2 = put_object_out_v2.version_id
    #     get_out = self.client.get_object(bucket_name, key)
    #     self.assertEqual(get_out.version_id, version_2)

    # def test_list_object_version(self):
    #     bucket_name = self.bucket_name + '-test-list-version'
    #     self.client.create_bucket(bucket_name)
    #     self.bucket_delete.append(bucket_name)
    #
    #     self.version_client.put_bucket_versioning(bucket_name, enable=True)
    #     for i in range(100):
    #         key = self.random_key('.js')
    #         content = random_bytes(1)
    #         self.client.put_object(bucket_name, key=key, content=content)
    #         self.client.put_object(bucket_name, key=key, content=random_bytes(1))
    #
    #     for i in range(10):
    #         self.client.put_object(bucket_name, key=str(i), content=random_bytes(1))
    #         self.client.put_object(bucket_name, key=str(i), content=random_bytes(1))
    #
    #     list_object_out = self.client.list_object_versions(bucket_name, max_keys=50, prefix=self.prefix)
    #     self.assertEqual(self.prefix, list_object_out.prefix)
    #     self.assertEqual(list_object_out.max_keys, 50)
    #     self.assertTrue(list_object_out.is_truncated)

    def test_append(self):
        bucket_name = self.bucket_name + '-test-append-object'
        key = self.random_key('.js')
        content = random_bytes(1024)

        self.client.create_bucket(bucket_name, az_redundancy=AzRedundancyType.Az_Redundancy_Multi_Az)
        self.bucket_delete.append(bucket_name)
        append_object_out = self.client.append_object(bucket_name, key, 0, content=content)
        self.assertTrue(append_object_out.hash_crc64_ecma > 0)
        self.assertEqual(append_object_out.next_append_offset, 1024)
        self.client.delete_object(bucket_name, key)

        key = self.random_key('.js')
        append_out_1 = self.client.append_object(bucket_name, key, 0, content=content[0:100], pre_hash_crc64_ecma=0)

        self.client.append_object(bucket_name, key, 100, content=content[100:],
                                  pre_hash_crc64_ecma=append_out_1.hash_crc64_ecma)

        get_out = self.client.get_object(bucket_name, key)
        self.assertEqual(get_out.read(), content)

    def test_append_with_options(self):
        bucket_name = self.bucket_name + '-test-append-with-options'
        key = self.random_key('.js')
        content = random_bytes(1024)

        def progress(consumed_bytes, total_bytes, rw_once_bytes,
                     type):
            print("consumed_bytes:{0},total_bytes{1}, rw_once_bytes:{2}, type:{3}".format(consumed_bytes, total_bytes,
                                                                                          rw_once_bytes, type))

        limiter = RateLimiter(5 * 1024 * 1024, 20 * 1024 * 1024)

        self.client.create_bucket(bucket_name, az_redundancy=AzRedundancyType.Az_Redundancy_Multi_Az)
        self.bucket_delete.append(bucket_name)
        meta = {'name': '张三', 'age': '13'}
        append_object_out = self.client.append_object(bucket_name, key, 0,
                                                      content=content,
                                                      content_length=1024,
                                                      cache_control="Cache-Control",
                                                      content_disposition="utf-8",

                                                      content_language="english",
                                                      content_encoding="utf-8",

                                                      expires=datetime.datetime.now(),
                                                      acl=ACLType.ACL_Private,
                                                      meta=meta,
                                                      website_redirect_location='/test',
                                                      data_transfer_listener=progress,
                                                      rate_limiter=limiter
                                                      )
        self.assertTrue(append_object_out.hash_crc64_ecma > 0)
        self.assertEqual(append_object_out.next_append_offset, 1024)
        get_object_out = self.client.get_object(bucket_name, key)
        self.assertTrue(get_object_out.object_type == 'Appendable')

    def test_list_object_info(self):
        bucket_name = self.bucket_name + '-test-list-object'
        key = self.random_key('.js')
        content = random_bytes(1024)

        self.client.create_bucket(bucket_name, az_redundancy=AzRedundancyType.Az_Redundancy_Multi_Az)
        self.bucket_delete.append(bucket_name)
        self.client.put_object(bucket_name, key=key, content=content)
        list_object_out = self.client.list_objects(bucket_name)
        self.assertEqual(list_object_out.name, bucket_name)
        self.assertFalse(list_object_out.is_truncated)
        self.assertEqual(list_object_out.max_keys, 1000)
        self.assertTrue(len(list_object_out.contents) > 0)

        object = list_object_out.contents[0]
        self.assertTrue(len(object.etag) > 0)
        self.assertTrue(len(object.key) > 0)
        self.assertTrue(object.last_modified is not None)
        self.assertTrue(object.size == 1024)
        self.assertTrue(len(object.owner.id) > 0)
        self.assertTrue(object.storage_class, StorageClassType.Storage_Class_Standard)

    def test_list_object_full_func(self):
        bucket_name = self.bucket_name + '-test-list-object'
        self.client.create_bucket(bucket_name, az_redundancy=AzRedundancyType.Az_Redundancy_Multi_Az)
        self.bucket_delete.append(bucket_name)

        for i in range(100):
            key = self.random_key('.js')
            content = random_bytes(1)
            self.client.put_object(bucket_name, key=key, content=content)

        for i in range(10):
            self.client.put_object(bucket_name, key=str(i), content=random_bytes(1))

        list_object_out = self.client.list_objects(bucket_name, max_keys=50, prefix=self.prefix)
        self.assertEqual(self.prefix, list_object_out.prefix)
        self.assertEqual(list_object_out.max_keys, 50)
        self.assertTrue(list_object_out.is_truncated)

        list_object_out_v2 = self.client.list_objects(bucket_name, max_keys=51, prefix=self.prefix,
                                                      marker=list_object_out.next_marker, delimiter=self.prefix,
                                                      reverse=False)
        self.assertEqual(len(list_object_out_v2.contents), 50)
        self.assertFalse(list_object_out_v2.is_truncated)

    def test_list(self):
        bucket_name = self.bucket_name + '-test-list-object'
        self.client.create_bucket(bucket_name, az_redundancy=AzRedundancyType.Az_Redundancy_Multi_Az)
        self.bucket_delete.append(bucket_name)
        self.client.put_object(bucket_name, 'test/')
        self.client.put_object(bucket_name, 'test/test')
        out = self.client.list_objects(bucket_name, prefix='test/', max_keys=1)

    def test_empty(self):
        bucket_name = self.bucket_name + '-test-empty'
        file_name = self.random_filename()
        self.client.create_bucket(bucket_name, acl=tos.ACLType.ACL_Public_Read_Write)
        self.bucket_delete.append(bucket_name)

        def percentage(consumed_bytes, total_bytes, rw_once_bytes,
                       type: DataTransferType):
            if total_bytes:
                rate = int(100 * float(consumed_bytes) / float(total_bytes))
                print("rate:{}, consumed_bytes:{},total_bytes{}, rw_once_bytes:{}, type:{}".format(rate, consumed_bytes,
                                                                                                   total_bytes,
                                                                                                   rw_once_bytes, type))

        rate_limiter = RateLimiter(rate=5 * 1024 * 1024, capacity=10 * 1024 * 1024)

        with open(file_name, 'wb+') as f:
            pass

        with open(file_name, 'rb+') as f:
            self.client.put_object(bucket_name, 'test', content=f, data_transfer_listener=percentage,
                                   rate_limiter=rate_limiter)
            self.assertEqual(self.client.get_object(bucket_name, 'test').read(), b'')

        rate_limiter = RateLimiter(rate=5 * 1024 * 1024, capacity=10 * 1024 * 1024)
        self.client.put_object(bucket_name, 'test123', content=StringIO(""))
        self.assertEqual(self.client.get_object(bucket_name, 'test123').read(), b'')
        self.client.put_object(bucket_name, 'test', content=BytesIO(b""))
        self.assertEqual(self.client.get_object(bucket_name, 'test').read(), b'')
        self.client.put_object(bucket_name, 'test', content="")
        self.assertEqual(self.client.get_object(bucket_name, 'test').read(), b'')
        self.client.put_object(bucket_name, 'test', content=b'')
        self.assertEqual(self.client.get_object(bucket_name, 'test').read(), b'')
        self.client.put_object_from_file(bucket_name, 'test', file_name, data_transfer_listener=percentage,
                                         rate_limiter=rate_limiter)
        self.assertEqual(self.client.get_object(bucket_name, 'test').read(), b'')
        self.client.upload_file(bucket_name, 'test', file_name, data_transfer_listener=percentage,
                                rate_limiter=rate_limiter)
        self.assertEqual(self.client.get_object(bucket_name, 'test').read(), b'')
        with self.assertRaises(TosClientError):
            out = self.client.append_object(bucket_name, 'test456', 0, StringIO(""))

        create_out = self.client.create_multipart_upload(bucket_name, 'test789')
        self.client.upload_part(bucket_name, 'test789', upload_id=create_out.upload_id, part_number=1,
                                content=StringIO(''))
        self.client.upload_part(bucket_name, 'test789', upload_id=create_out.upload_id, part_number=1,
                                content=BytesIO(b''))
        self.client.upload_part_from_file(bucket_name, 'test789', upload_id=create_out.upload_id, part_number=1,
                                          file_path=file_name)

    def test_list_object_with_case(self):
        bucket_name = self.bucket_name + '-test-list-object-with-case'
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        for i in range(100):
            key = self.random_key('.js')
            content = random_string(1000)
            self.client.put_object(bucket_name, key, content=content)

        is_truncated = True
        count = 0
        object_count = 0
        marker = ''
        while is_truncated:
            if count > 10:
                break
            count += 1
            list_out = self.client.list_objects(bucket_name, max_keys=10, marker=marker)
            is_truncated = list_out.is_truncated
            marker = list_out.next_marker
            object_count += len(list_out.contents)
        self.assertEqual(count, 10)
        self.assertEqual(object_count, 100)

        deletes = []
        for i in range(3):
            for j in range(3):
                for k in range(3):
                    path = '{}/{}/{}'.format(i, j, k)
                    deletes.append(Deleted(key=path))
                    self.client.put_object(bucket_name, path, content=b'')

        self.client.list_objects(bucket_name, prefix='0')
        self.client.list_objects(bucket_name, prefix='1')
        self.client.list_objects(bucket_name, prefix='0/1')

        out4 = self.client.list_objects(bucket_name, delimiter='/')

        def dfs(list_out: ListObjectsOutput):
            if len(list_out.common_prefixes) == 0:
                for e in list_out.contents:
                    print(e)
                return len(list_out.contents)
            else:
                a = 0
                for prefix in list_out.common_prefixes:
                    a += dfs(self.client.list_objects(bucket_name, delimiter='/', prefix=prefix.prefix))
                return a

        count = dfs(out4)
        self.assertEqual(count, 27)

        self.client.delete_multi_objects(bucket_name, deletes)

        out5 = self.client.list_objects(bucket_name, delimiter='/')
        self.assertEqual(len(out5.common_prefixes), 0)

    def test_set_object_meta(self):
        bucket_name = self.bucket_name + '-test-set-object-meta'
        key = self.random_key('.js')
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        meta = {'name': 'sunyushan', 'age': '10'}
        self.client.put_object(bucket_name, key=key, content=random_bytes(10), meta=meta)
        meta['name'] = '张三'
        self.client.set_object_meta(bucket_name, key, meta=meta)

        get_object_out = self.client.get_object(bucket_name, key=key)
        self.assertEqual(meta['name'], get_object_out.meta['name'])

    def test_get_object_meta(self):
        bucket_name = self.bucket_name + '-test-get-object-meta'
        key = self.random_key('.js')
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        meta = {'name': 'jason', 'age': '10'}
        self.client.put_object(bucket_name, key=key, content=random_bytes(10), meta=meta)

        self.client.get_object(bucket_name, key=key)

    def test_get_object_with_data_transfer_listener(self):
        bucket_name = self.bucket_name + '-test-with-transfer-listener'
        key = self.random_key('.js')
        content = random_bytes(1025 * 1024)
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)

        def progress(consumed_bytes, total_bytes, rw_once_bytes,
                     type: DataTransferType):
            print(
                "consumed_bytes:{0},total_bytes{1}, rw_once_bytes:{2}, type:{3}".format(consumed_bytes, total_bytes,
                                                                                        rw_once_bytes, type))

        self.client.put_object(bucket_name, key=key, content=content)

        get_object_out = self.client.get_object(bucket_name, key=key, data_transfer_listener=progress,
                                                rate_limiter=RateLimiter(1024 * 1024 * 5, 1024 * 1024 * 20))
        self.assertEqual(get_object_out.read(), content)
        # self.assertEqual(get_object_out.hash_crc64_ecma, get_object_out.client_crc)

    def test_get_object_to_file(self):
        bucket_name = self.bucket_name + 'download-dir'
        cwd = os.getcwd()
        key = self.random_key('.txt')
        content = random_bytes(1024 * 1024 * 5)
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        file_name = self.random_filename()
        with open(file_name, 'wb') as f:
            f.write(content)
        self.client.put_object_from_file(bucket_name, key, file_name)
        self.client.get_object_to_file(bucket_name, key, cwd + '/test.txt')
        self.assertFileContent(cwd + '/test.txt', content)

        content = content + b'1'
        self.client.put_object(bucket_name, key, content=content)
        self.client.get_object_to_file(bucket_name, key, cwd + '/test.txt')
        self.assertFileContent(cwd + '/test.txt', content)
        os.remove(cwd + '/test.txt')

        self.client.get_object_to_file(bucket_name, key, cwd + '/test/f1')
        self.assertFileContent(cwd + '/test/f1', content)

        self.client.get_object_to_file(bucket_name, key, cwd + '/dir1/')
        self.assertFileContent(cwd + '/dir1/' + key, content)
        os.remove(cwd + '/dir1/' + key)

        key = key + '/'
        self.client.put_object(bucket_name, key, content=content)
        self.client.get_object_to_file(bucket_name, key, cwd + '/dir3/')
        self.assertTrue(os.path.isdir(cwd + '/dir3/' + key))

    def test_with_rate_limiter(self):
        bucket_name = self.bucket_name + '-test-with-rate-limiter'
        key = self.random_key('.js')
        content = random_bytes(1024 * 1024)
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)

        def progress(consumed_bytes, total_bytes, rw_once_bytes,
                     type: DataTransferType):
            print(
                "consumed_bytes:{0},total_bytes{1}, rw_once_bytes:{2}, type:{3}".format(consumed_bytes, total_bytes,
                                                                                        rw_once_bytes, type))

        limiter = RateLimiter(1024 * 1024 * 5, 1024 * 1024 * 20)
        self.client.put_object(bucket_name, key=key, content=content)

        get_object_out = self.client.get_object(bucket_name, key=key)

        self.client.put_object(bucket_name, key + '1', content=get_object_out, data_transfer_listener=progress,
                               rate_limiter=limiter)
        get_object_out = self.client.get_object(bucket_name, key=key, data_transfer_listener=progress,
                                                rate_limiter=limiter)
        self.assertEqual(get_object_out.read(), content)

    def test_put_object_acl(self):
        bucket_name = self.bucket_name + '-test-put-object-acl'
        key = self.random_key('.js')
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        self.client.put_object(bucket_name, key, content=random_bytes(5))
        grants = []
        grantee = Grantee(id="123", display_name="123", type=GranteeType.Grantee_User)
        grant = Grant(grantee, permission=PermissionType.Permission_Full_Control)
        grants.append(grant)
        # self.client.put_object_acl(bucket_name, key, acl=ACLType.ACL_Bucket_Owner_Full_Control)
        # self.client.get_object_acl(bucket_name, key)
        self.client.put_object_acl(bucket_name, key, owner=Owner("123", "test"), grants=grants)

    def test_put_with_chunked_invalid_length(self):
        bucket_name = self.bucket_name + '-put-with-chunked-invalid-length'
        key = self.random_key('.js')
        data = random_bytes(5)
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        self.client.put_object(bucket_name, key, content=data)

        get_out = self.client.get_object(bucket_name, key)
        key2 = self.random_key('.js')
        self.client.put_object(bucket_name, key2, content=get_out.content)

        get_out = self.client.get_object(bucket_name, key)
        self.client.put_object(bucket_name, key2, content=get_out.content, content_length=get_out.content_length)

        get_out = self.client.get_object(bucket_name, key2)
        self.assertEqual(get_out.read(), data)

        get_out = self.client.get_object(bucket_name, key2)
        create_out = self.client.create_multipart_upload(bucket_name, key2)
        self.client.upload_part(bucket_name, key2, upload_id=create_out.upload_id, part_number=1,
                                content=get_out.content, content_length=get_out.content_length)

        get_out = self.client.get_object(bucket_name, key2)
        self.client.upload_part(bucket_name, key2, upload_id=create_out.upload_id, part_number=2,
                                content=get_out.content)

    def test_anonymous(self):
        bucket_name = self.bucket_name + '-anonymous'
        key = self.random_key('.js')
        data = random_bytes(5)
        self.client.create_bucket(bucket_name, acl=ACLType.ACL_Public_Read_Write)
        self.client.put_object(bucket_name, key, acl=ACLType.ACL_Public_Read_Write, content=data)

        client = TosClientV2("", "", self.endpoint, self.region)
        out = client.get_object(bucket_name, key)
        self.assertEqual(out.read(), data)

    def test_put_with_md5(self):
        bucket_name = self.bucket_name + '-put-with-md5'
        self.bucket_delete.append(bucket_name)
        key = self.random_key('.js')
        content = random_bytes(100)
        file_name = random_string(5)
        with open(file_name, 'wb') as f:
            f.write(content)

        md5 = 0
        with open(file_name, 'rb') as f:
            md5 = calculate_md5(f)

        self.client.create_bucket(bucket_name)

        self.client.put_object_from_file(bucket_name, key, file_name, content_md5=md5.decode('utf-8'))

        with self.assertRaises(TosServerError):
            self.client.put_object_from_file(bucket_name, key, file_name, content_md5='test_error')
        os.remove(path=file_name)

    def test_tagging(self):
        bucket_name = self.bucket_name + 'tagging'
        key = self.random_key('.js')
        content = random_bytes(100)
        self.client.create_bucket(bucket_name)
        self.client.put_object(bucket=bucket_name, key=key, content=content)
        tag_set = []
        tag_set.append(Tag(
            key='1',
            value='2'
        ))
        tag_set.append(Tag(
            key='3',
            value='4'
        ))
        put_out = self.client.put_object_tagging(bucket=bucket_name, key=key, tag_set=tag_set)
        self.assertIsNone(put_out.version_id)

        get_out = self.client.get_object_tagging(bucket=bucket_name, key=key)
        self.assertIsNotNone(get_out.request_id)
        get_set = get_out.tag_set
        self.assertTrue(len(get_set) == 2)
        self.assertEqual(get_set[0].key, '1')
        self.assertEqual(get_set[0].value, '2')

        self.assertEqual(get_set[1].key, '3')
        self.assertEqual(get_set[1].value, '4')

        delete_out = self.client.delete_object_tagging(bucket=bucket_name, key=key)
        self.assertIsNone(delete_out.version_id)
        get_out_2 = self.client.get_object_tagging(bucket=bucket_name, key=key)
        self.assertTrue(len(get_out_2.tag_set) == 0)

        self.client.delete_object(bucket=bucket_name, key=key)
        self.client.delete_bucket(bucket_name)

    def test_list_object_v2(self):
        bucket_name = self.bucket_name + '-test-list-object-with-case'
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        for i in range(3):
            for j in range(3):
                for k in range(3):
                    path = '{}/{}/{}'.format(i, j, k)
                    self.client.put_object(bucket_name, path, content=b'1')
        self.client.put_object(bucket_name, 'key')

        # out_2 = self.client.list_objects_type2(bucket=bucket_name, prefix='0', start_after='0/1', max_keys=2)
        # out_2_reverse = self.client.list_objects_type2(bucket=bucket_name, prefix='0', start_after='0/1', max_keys=2)
        # out_3 = self.client.list_objects_type2(bucket=bucket_name, prefix='0', start_after='0/1', max_keys=2,
        #                                        delimiter='/', continuation_token=out_2.next_continuation_token)
        #

        out_base = self.client.list_objects_type2(bucket_name, delimiter='/', max_keys=1)
        self.assertEqual(out_base.name, bucket_name)
        self.assertIsNotNone(out_base.request_id)
        self.assertTrue(out_base.is_truncated)
        self.assertEqual(out_base.delimiter, '/')
        self.assertEqual(out_base.max_keys, 1)
        self.assertIsNotNone(out_base.next_continuation_token)
        out_base_2 = self.client.list_objects_type2(bucket_name, continuation_token=out_base.next_continuation_token)
        self.assertIsNotNone(out_base_2.continuation_token)

        out_base_3 = self.client.list_objects_type2(bucket_name, prefix='0', max_keys=1)
        self.assertEqual(out_base_3.prefix, '0')

        continuation_token = None
        is_truncated = True
        count = 0
        while is_truncated:
            out = self.client.list_objects_type2(bucket_name, continuation_token=continuation_token, max_keys=2,
                                                 delimiter='/')
            is_truncated = out.is_truncated
            continuation_token = out.next_continuation_token
            keycount = len(out.contents) + len(out.common_prefixes)
            self.assertEqual(keycount, out.key_count)
            count += keycount
            if is_truncated:
                self.assertIsNotNone(continuation_token)
        self.assertEqual(count, 4)

        continuation_token = None
        is_truncated = True
        count = 0
        while is_truncated:
            out = self.client.list_objects_type2(bucket_name, continuation_token=continuation_token, max_keys=2)
            is_truncated = out.is_truncated
            continuation_token = out.next_continuation_token
            keycount = len(out.contents) + len(out.common_prefixes)
            self.assertEqual(keycount, out.key_count)
            self.assertIsNotNone(out.contents)
            for c in out.contents:
                self.assertTrue(c.owner.id)
                self.assertTrue(c.owner.display_name)
            count += keycount
            if is_truncated:
                self.assertIsNotNone(continuation_token)
        self.assertEqual(count, 28)

        continuation_token = None
        is_truncated = True
        count = 0
        while is_truncated:
            out = self.client.list_objects_type2(bucket_name, continuation_token=continuation_token)
            is_truncated = out.is_truncated
            continuation_token = out.next_continuation_token
            keycount = len(out.contents) + len(out.common_prefixes)
            self.assertEqual(keycount, out.key_count)
            count += keycount
            if is_truncated:
                self.assertIsNotNone(continuation_token)
        self.assertEqual(count, 28)

        continuation_token = None
        is_truncated = True
        count = 0
        while is_truncated:
            out = self.client.list_objects_type2(bucket_name, continuation_token=continuation_token,
                                                 delimiter='/')
            is_truncated = out.is_truncated
            continuation_token = out.next_continuation_token
            keycount = len(out.contents) + len(out.common_prefixes)
            self.assertEqual(keycount, out.key_count)
            count += keycount
            if is_truncated:
                self.assertIsNotNone(continuation_token)
        self.assertEqual(count, 4)

    def test_fetch_object(self):
        bucket_name = self.bucket_name + '-fetch-object'
        bucket_fetch = self.bucket_name + '-fetch-test'
        self.bucket_delete.append(bucket_name)
        self.bucket_delete.append(bucket_fetch)
        object_name = 'test.txt'
        self.client.create_bucket(bucket_fetch)
        self.client.put_object(bucket_fetch, object_name)
        self.client.put_object_acl(bucket=bucket_fetch, key=object_name, acl=ACLType.ACL_Public_Read_Write)
        key = self.random_key('.js')
        self.client.create_bucket(bucket=bucket_name)
        meta = {'姓名': '张三'}
        fetch_out = self.client.fetch_object(bucket=bucket_name, key=key,
                                             url="https://{}.{}".format(bucket_fetch,
                                                                        self.endpoint) + '/' + object_name,
                                             meta=meta, acl=ACLType.ACL_Public_Read)
        out = self.client.get_object(bucket=bucket_name, key=key)
        acl_out = self.client.get_object_acl(bucket_name, key)
        self.assertEqual(acl_out.grants[0].permission, PermissionType.Permission_Read)
        get_out = self.client.get_object(bucket=bucket_name, key=key)
        get_out.meta['姓名'] = meta['姓名']

    def test_fetch_task_object(self):
        bucket_name = self.bucket_name + '-fetch-object'
        self.bucket_delete.append(bucket_name)
        bucket_fetch = self.bucket_name + '-fetch-task-test'
        self.bucket_delete.append(bucket_fetch)

        key = self.random_key('123')

        self.client.create_bucket(bucket=bucket_fetch)
        self.client.put_object(bucket=bucket_fetch, key=key, acl=ACLType.ACL_Public_Read_Write)
        url = 'http://{}.{}'.format(bucket_fetch, self.endpoint) + '/' + key
        self.client.create_bucket(bucket=bucket_name)
        out = self.client.put_fetch_task(bucket=bucket_name, key=key, url=url)
        self.assertIsNotNone(out.task_id)
        time.sleep(10)

    def test_post_object(self):
        bucket_name = self.bucket_name + '-post-object'
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        file_name = self.random_filename()
        content = random_bytes(1024)
        with open(file_name, 'wb+') as f:
            f.write(content)
        key = self.random_key()
        out = self.client.pre_signed_post_signature(bucket=bucket_name, key=key, conditions=[])
        form = {'key': key, 'x-tos-algorithm': out.algorithm, 'bucket': bucket_name, 'x-tos-date': out.date,
                'policy': out.policy, 'x-tos-signature': out.signature, 'x-tos-credential': out.credential}
        resp = requests.post(url=self.client._make_virtual_host_url(bucket_name, key),
                             files={"upload_file": open(file_name, 'rb')},
                             data=form)
        self.assertEqual(resp.status_code, 204)
        resp.close()

        condition = [PostSignatureCondition(key='x-tos-acl', value='private', operator='eq')]
        out2 = self.client.pre_signed_post_signature(conditions=condition, bucket=bucket_name, key=key)
        form2 = {'key': key, 'x-tos-algorithm': out2.algorithm, 'bucket': bucket_name, 'x-tos-date': out2.date,
                 'policy': out2.policy, 'x-tos-signature': out2.signature, 'x-tos-credential': out2.credential,
                 'x-tos-acl': 'private'}
        resp = requests.post(url=self.client._make_virtual_host_url(bucket_name),
                             files={"upload_file": open(file_name, 'rb')},
                             data=form2)
        resp.close()

        self.assertEqual(resp.status_code, 204)

        form2['x-tos-acl'] = 'public-read'
        resp = requests.post(url=self.client._make_virtual_host_url(bucket_name),
                             files={"upload_file": open(file_name, 'rb')},
                             data=form2)
        resp.close()
        self.assertEqual(resp.status_code, 403)

        prefix = 'test_post/'
        key = prefix + key
        condition = [PostSignatureCondition(key='key', value=prefix, operator='starts-with')]
        out3 = self.client.pre_signed_post_signature(conditions=condition, bucket=bucket_name)
        form3 = {'key': key, 'x-tos-algorithm': out3.algorithm, 'bucket': bucket_name, 'x-tos-date': out3.date,
                 'policy': out3.policy, 'x-tos-signature': out3.signature, 'x-tos-credential': out3.credential}
        resp = requests.post(url=self.client._make_virtual_host_url(bucket_name, key),
                             files={"upload_file": open(file_name, 'rb')},
                             data=form3)
        self.assertEqual(resp.status_code, 204)

        resp = self.client.get_object(bucket_name, key)
        self.assertEqual(resp.status_code, 200)

    def test_pre_signed_policy_url(self):
        bucket_name = self.bucket_name + 'test-policy-url'
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        self.client.put_bucket_versioning(bucket_name, VersioningStatusType.Versioning_Status_Enabled)
        time.sleep(60)
        self.client.put_object(bucket_name, 'abc/')
        self.client.put_object(bucket_name, 'abc/abc/')
        out_v1 = self.client.put_object(bucket_name, 'exampleobject', content=b'1')
        out_v2 = self.client.put_object(bucket_name, 'exampleobject', content=b'3')
        self.client.put_object(bucket_name, 'exampleobject1', content=b'2')
        conditions = [PolicySignatureCondition(key='key', value='abc/', operator='starts-with'),
                      # PostSignatureCondition(key='key', value='abc/abc/', operator='starts-with'),
                      PolicySignatureCondition(key='key', value='exampleobject', operator='eq'),
                      PolicySignatureCondition(key='key', value='exampleobject1')]
        out = self.client.pre_signed_policy_url(bucket_name, conditions)
        o1_2 = out.get_signed_url_for_list()
        o2_2 = out.get_signed_url_for_get_or_head('test')
        list_url = out.get_signed_url_for_list({'prefix': 'abc/abc/'})
        get_url = out.get_signed_url_for_get_or_head(key='exampleobject',
                                                     additional_query={'versionId': out_v1.version_id})
        list_ans = requests.get(list_url)
        self.assertTrue(b'abc/' in list_ans.content)
        get_out = requests.get(get_url)
        self.assertEqual(b'1', get_out.content)

        out_1 = self.client.pre_signed_policy_url(bucket_name, conditions,
                                                  alternative_endpoint='tos-cn-beijing.volces.com')
        self.assertEqual(out_1._host, 'tos-cn-beijing.volces.com')
        self.assertEqual(out_1._scheme, 'https://')
        out_11 = self.client.pre_signed_policy_url(bucket_name, conditions,
                                                   alternative_endpoint='tos-cn-beijing.volces.com',
                                                   is_custom_domain=True)
        self.assertIsNone(out_11._bucket)
        out_2 = self.client.pre_signed_policy_url(bucket_name, conditions,
                                                  alternative_endpoint='http://tos-cn-beijing.volces.com')
        self.assertEqual(out_2._host, 'tos-cn-beijing.volces.com')
        self.assertEqual(out_2._scheme, 'http://')
        out_3 = self.client.pre_signed_policy_url(bucket_name, conditions,
                                                  alternative_endpoint='https://tos-cn-beijing.volces.com')
        self.assertEqual(out_3._host, 'tos-cn-beijing.volces.com')
        self.assertEqual(out_3._scheme, 'https://')

    def test_wrapper_socket_io(self):
        bucket_name = self.bucket_name + 'test-wrapper-crc'
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        case = [True, False]
        for i in case:
            for j in case:
                for k in case:
                    print('init:{}, crc:{} use_data_transfer_listener: {}, ues_limiter: {}'.format("yes", i, j, k))
                    self.wrapper_socket_io(True, i, j, k, bucket_name)

    def test_wrapper(self):
        bucket_name = self.bucket_name + 'test-wrapper-crc'
        content = random_bytes(1024 * 1024 * 6)
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        case = [True, False]
        # for i in case:
        #     for j in case:
        #         for k in case:
        #             print('init:{}, crc:{} use_data_transfer_listener: {}, ues_limiter: {}'.format("yes", i, j, k))
        #             self.wappper(True, i, j, k, bucket_name, content, reset_content=content)

        file_name = self.random_filename()
        with open(file_name, 'wb') as f:
            f.write(content)
        with open(file_name, 'rb') as f:
            size = os.path.getsize(file_name)
            body = tos.utils._make_upload_part_file_content(f, offset=1024, part_size=1024, size=size)
            for i in case:
                for j in case:
                    for k in case:
                        print('init:{}, crc:{} use_data_transfer_listener: {}, ues_limiter: {}'.format("yes", i, j, k))
                        self.wappper(True, i, j, k, bucket_name, content[1024:2048], reset_content=content[1024:2048])

    def test_get_object_range(self):
        bucket_name = self.bucket_name + 'object-range'
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        key = self.random_key('.js')
        key_copy = self.random_key('.js')
        content = random_bytes(1024 * 1024)
        self.client.put_object(bucket_name, key, content=content)
        out = self.client.get_object(bucket_name, key, range='bytes=0-1023')
        self.assertEqual(out.read(), content[0:1024])
        out_2 = self.client.get_object(bucket_name, key, range='bytes=1024-')
        self.assertEqual(out_2.read(), content[1024:])

    def test_copy_object_range(self):
        src_bucket_name = self.bucket_name + 'copy-object-range'
        key = self.random_key('.js')
        save_bucket_name = 'save'
        content = random_bytes(1024)
        self.client.create_bucket(src_bucket_name)
        self.client.create_bucket(save_bucket_name)
        self.bucket_delete.append(save_bucket_name)
        self.bucket_delete.append(src_bucket_name)

        self.client.put_object(bucket=src_bucket_name, key=key, content=content)

        out = self.client.create_multipart_upload(save_bucket_name, key)
        parts = []
        part_copy_1 = self.client.upload_part_copy(out.bucket, out.key, out.upload_id, part_number=1,
                                                   src_bucket=src_bucket_name, src_key=key,
                                                   copy_source_range='bytes=0-100')

        parts.append(UploadedPart(1, part_copy_1.etag))

        self.client.complete_multipart_upload(save_bucket_name, key, out.upload_id, parts)

        out = self.client.get_object(bucket=save_bucket_name, key=key)
        self.assertEqual(out.read(), content[0:101])

    def test_non_file(self):
        bucket_name = self.bucket_name + 'non-file'
        key = self.random_key('.js')
        file_name = self.random_filename()
        download_file_1 = self.random_filename()
        download_file_2 = self.random_filename()
        with open(file_name, 'wb+') as f:
            pass
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        self.client.put_object_from_file(bucket_name, key, file_name)
        self.client.head_object(bucket_name, key)
        self.client.get_object_to_file(bucket_name, key, download_file_1)

        self.client.delete_object(bucket_name, key)
        self.client.upload_file(bucket_name, key, file_name)

        self.client.head_object(bucket_name, key)
        self.client.download_file(bucket_name, key, download_file_2)

        # 创建 TosClientV2 对象，对桶和对象的操作都通过 TosClientV2 实现
        def copy_event(copy_event_type: CopyEventType, err, bucket, key, upload_id, src_bucket, src_key, src_version_id,
                       checkpoint, copy_part):
            print(copy_event_type, err, bucket, key, upload_id, src_bucket, src_key, src_version_id, checkpoint,
                  copy_part)

        resumable_copy_object = self.client.resumable_copy_object(bucket_name, key + '1', bucket_name, key,
                                                                  copy_event_listener=copy_event)
        head_out = self.client.head_object(bucket=bucket_name, key=key + '1')
        get_out = self.client.get_object(bucket=bucket_name, key=key + '1')
        assert get_out.read() == b''
        self.client.copy_object(bucket_name, key + '1', bucket_name, key)

        out = self.client.create_multipart_upload(bucket_name, key + '2')
        part = self.client.upload_part_from_file(bucket_name, key + '2', upload_id=out.upload_id, part_number=1,
                                                 file_path=file_name)
        self.client.complete_multipart_upload(bucket_name, key + '2', upload_id=out.upload_id, parts=[part])
        self.client.head_object(bucket_name, key + '2')

    def test_restore_object(self):
        bucket_name = self.bucket_name + 'coldarchive'
        content = random_bytes(1024)
        key = self.random_key('.js')
        self.client.create_bucket(bucket=bucket_name, storage_class=StorageClassType.Storage_Class_Cold_Archive)
        self.bucket_delete.append(bucket_name)

        head_cold = self.client.head_bucket(bucket_name)
        self.assertEqual(head_cold.storage_class, StorageClassType.Storage_Class_Cold_Archive)

        obj = self.client.put_object(bucket=bucket_name, key=key)
        head_1 = self.client.head_object(bucket=bucket_name, key=key)
        self.assertEqual(head_1.storage_class, StorageClassType.Storage_Class_Cold_Archive)

        self.client.put_object(bucket=bucket_name, key=key, content=content)
        resp = self.client.restore_object(bucket=bucket_name, key=key, days=1,
                                          restore_job_parameters=RestoreJobParameters(TierType.Tier_Expedited))

        head_2 = self.client.head_object(bucket=bucket_name, key=key)
        self.assertEqual(head_2.storage_class, StorageClassType.Storage_Class_Cold_Archive)
        head_2.restore = 'ongoing-request="true"'
        head_2.restore_tier = TierType.Tier_Standard
        head_2.restore_expiry_days = 1
        time.sleep(60 * 5)

        out = self.client.get_object(bucket=bucket_name, key=key)
        self.assertEqual(out.read(), content)
        assert 'ongoing-request="false"' in out.restore

        self.client.put_object(bucket_name, key, storage_class=StorageClassType.Storage_Class_Archive)
        head_out = self.client.head_object(bucket_name, key)
        assert head_out.storage_class == StorageClassType.Storage_Class_Archive

    def test_traffic_limit(self):
        bucket_name = self.bucket_name + 'test-traffic-limit'
        content = random_bytes(MIN_TRAFFIC_LIMIT * 5)
        key = self.random_key('.js')
        self.client.create_bucket(bucket_name)
        start = time.time()
        self.client.put_object(bucket_name, key, content=content, traffic_limit=MIN_TRAFFIC_LIMIT)
        end = time.time()
        assert end - start > 5

    def test_rename_object(self):
        bucket_name = self.bucket_name + '-rename'
        content = random_bytes(1024)
        key = self.random_key('.js')
        self.client.create_bucket(bucket=bucket_name)
        self.bucket_delete.append(bucket_name)

        self.client.put_bucket_rename(bucket=bucket_name, rename_enable=True)
        time.sleep(30)
        bucket_rename_output = self.client.get_bucket_rename(bucket=bucket_name)
        self.assertEqual(bucket_rename_output.rename_enable, True)

        new_key = self.random_key('.js')
        self.client.put_object(bucket=bucket_name, key=key, content=content)
        self.client.rename_object(bucket=bucket_name, key=key, new_key=new_key)
        with self.assertRaises(TosServerError) as cm:
            self.client.get_object(bucket=bucket_name, key=key)
        self.assertEqual(cm.exception.status_code, 404)

        get_object_out = self.client.get_object(bucket=bucket_name, key=new_key)
        self.assertEqual(get_object_out.read(), content)

        self.client.delete_bucket_rename(bucket=bucket_name)
        bucket_rename_output = self.client.get_bucket_rename(bucket=bucket_name)
        self.assertEqual(bucket_rename_output.rename_enable, False)

    def test_callback(self):
        bucket_name = self.bucket_name + '-callback'
        content = random_bytes(1024)
        key = self.random_key('.js')
        self.client.create_bucket(bucket=bucket_name)
        self.bucket_delete.append(bucket_name)

        callback_url = '{"callbackUrl" : "http://www.test.xxx.com"}'
        callback = base64.b64encode(callback_url.encode('utf-8')).decode('utf-8')
        with self.assertRaises(TosServerError) as cm:
            self.client.put_object(bucket=bucket_name, key=key, content=content, callback=callback)
        self.assertEqual(cm.exception.status_code, 203)

        callback = base64.b64encode(self.callback.encode('utf-8')).decode('utf-8')
        callback_var = base64.b64encode(self.callback_var.encode('utf-8')).decode('utf-8')
        out = self.client.put_object(bucket=bucket_name, key=key, content=content, callback=callback,
                                     callback_var=callback_var)
        self.assertEqual(out.callback_result, '{"msg":"ok"}')

    def test_custom_domain(self):
        bucket_name = self.bucket_name + '-custom-domain'
        self.client.create_bucket(bucket=bucket_name)
        self.bucket_delete.append(bucket_name)
        content = random_bytes(100)
        key = self.random_key('.js')
        schema, host = _get_host_schema(self.endpoint)
        endpoint = schema + bucket_name + '.' + host

        client = TosClientV2(self.ak, self.sk, endpoint, self.region, enable_crc=True, max_retry_count=2,
                             is_custom_domain=True)
        client.put_object(bucket=bucket_name, key=key, content=content)
        get_out = self.client.get_object(bucket=bucket_name, key=key)
        self.assertTrue(get_out.read(), content)

        signed_url_out = client.pre_signed_url(HttpMethodType.Http_Method_Get, bucket=bucket_name, key=key)
        rsp = requests.get(signed_url_out.signed_url)
        self.assertEqual(rsp.content, content)

        client = TosClientV2(self.ak, self.sk, endpoint, self.region, enable_crc=True, max_retry_count=2)
        signed_url_out = client.pre_signed_url(HttpMethodType.Http_Method_Get, bucket=bucket_name, key=key,
                                               is_custom_domain=True)
        rsp = requests.get(signed_url_out.signed_url)
        self.assertEqual(rsp.content, content)

    def test_env_credential_provider(self):
        bucket_name = self.bucket_name + '-env-credential-provider'
        os.environ['TOS_ACCESS_KEY'] = self.ak
        os.environ['TOS_SECRET_KEY'] = self.sk
        client = TosClientV2(region=self.region, endpoint=self.endpoint, credentials_provider=EnvCredentialsProvider())
        client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        key = self.random_key('.js')
        client.put_object(bucket_name, key=key)
        client.delete_object(bucket_name, key=key)

    def wrapper_socket_io(self, init, crc, use_data_transfer_listener, ues_limiter, bucket_name):
        def progress(consumed_bytes, total_bytes, rw_once_bytes,
                     data_type: DataTransferType):
            print("consumed_bytes:{0},total_bytes{1}, rw_once_bytes:{2}, type:{3}".format(consumed_bytes, total_bytes,
                                                                                          rw_once_bytes, data_type))

        data_transfer_listener = None
        limiter = None
        if use_data_transfer_listener:
            data_transfer_listener = progress
        if ues_limiter:
            limiter = RateLimiter(1024 * 1024, 1024 * 1024 * 5)
        if crc:
            client = tos.TosClientV2(self.ak, self.sk, self.endpoint, self.region)
        else:
            client = tos.TosClientV2(self.ak, self.sk, self.endpoint, self.region, enable_crc=False)
        key = self.random_key('.js')
        content = get_socket_io()
        if init:
            content = tos.utils.init_content(content)
            self.assertTrue(isinstance(content, tos.utils._IterableAdapter))
        if data_transfer_listener:
            content = tos.utils.add_progress_listener_func(content, data_transfer_listener)
            self.assertTrue(isinstance(content, tos.utils._IterableAdapter))
        if limiter:
            content = tos.utils.add_rate_limiter_func(content, limiter)
            self.assertTrue(isinstance(content, tos.utils._IterableAdapter))
        if crc:
            content = tos.utils.add_crc_func(content)
            self.assertTrue(isinstance(content, tos.utils._IterableAdapter))

        self.assertTrue(hasattr(content, 'can_reset'))
        self.assertEqual(content.can_reset, False)
        out = client.put_object(bucket_name, key=key, content=content)
        if crc:
            self.assertEqual(out.hash_crc64_ecma, content.crc)
        # self.assertObjectContent(bucket_name, key, get_socket_io().read())

    def wappper(self, init, crc, use_data_transfer_listener, ues_limiter, bucket_name, content, reset_content=None):

        def progress(consumed_bytes, total_bytes, rw_once_bytes,
                     data_type: DataTransferType):
            print("consumed_bytes:{0},total_bytes{1}, rw_once_bytes:{2}, type:{3}".format(consumed_bytes, total_bytes,
                                                                                          rw_once_bytes, data_type))

        origin = content
        data_transfer_listener = None
        limiter = None
        if use_data_transfer_listener:
            data_transfer_listener = progress
        if ues_limiter:
            limiter = RateLimiter(1024 * 1024 * 3, 1024 * 1024 * 10)
        if crc:
            client = tos.TosClientV2(self.ak, self.sk, self.endpoint, self.region)
        else:
            client = tos.TosClientV2(self.ak, self.sk, self.endpoint, self.region, enable_crc=False)
        key = self.random_key('.js')
        if init:
            content = tos.utils.init_content(content)
            self.assertTrue(isinstance(content, tos.utils._ReaderAdapter))
            self.assertEqual(content.can_reset, True)
        if data_transfer_listener:
            content = tos.utils.add_progress_listener_func(content, data_transfer_listener)
            self.assertTrue(isinstance(content, tos.utils._ReaderAdapter))
            self.assertEqual(content.can_reset, True)
        if limiter:
            content = tos.utils.add_rate_limiter_func(content, limiter)
            self.assertTrue(isinstance(content, tos.utils._ReaderAdapter))
            self.assertEqual(content.can_reset, True)
        if crc:
            content = tos.utils.add_crc_func(content)
            self.assertTrue(isinstance(content, tos.utils._ReaderAdapter))
            self.assertEqual(content.can_reset, True)

        self.assertTrue(hasattr(content, 'can_reset'))
        self.assertEqual(content.can_reset, True)
        out = client.put_object(bucket_name, key=key, content=content)
        if limiter:
            self.assertTrue(limiter._current_amount > 0)
        if crc:
            self.assertEqual(out.hash_crc64_ecma, content.crc)
        self.assertObjectContent(bucket_name, key, origin)

        if reset_content:
            content.reset()
            client.put_object(bucket_name, key, content=content)
            if crc:
                self.assertEqual(out.hash_crc64_ecma, content.crc)
            self.assertObjectContent(bucket_name, key, reset_content)

    def random_key(self, suffix=''):
        key = self.prefix + random_string(12) + suffix
        return key


if __name__ == '__main__':
    unittest.main()
